# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Session hook for logging benchmark metric."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf  # pylint: disable=g-bad-import-order


class LoggingMetricHook(tf.estimator.LoggingTensorHook):
  """Hook to log benchmark metric information.

  This hook is very similar as tf.train.LoggingTensorHook, which logs given
  tensors every N local steps, every N seconds, or at the end. The metric
  information will be logged to given log_dir or via metric_logger in JSON
  format, which can be consumed by data analysis pipeline later.

  Note that if `at_end` is True, `tensors` should not include any tensor
  whose evaluation produces a side effect such as consuming additional inputs.
  """

  def __init__(self, tensors, metric_logger=None,
               every_n_iter=None, every_n_secs=None, at_end=False):
    """Initializer for LoggingMetricHook.

    Args:
      tensors: `dict` that maps string-valued tags to tensors/tensor names,
          or `iterable` of tensors/tensor names.
      metric_logger: instance of `BenchmarkLogger`, the benchmark logger that
          hook should use to write the log.
      every_n_iter: `int`, print the values of `tensors` once every N local
          steps taken on the current worker.
      every_n_secs: `int` or `float`, print the values of `tensors` once every N
          seconds. Exactly one of `every_n_iter` and `every_n_secs` should be
          provided.
      at_end: `bool` specifying whether to print the values of `tensors` at the
          end of the run.

    Raises:
      ValueError:
        1. `every_n_iter` is non-positive, or
        2. Exactly one of every_n_iter and every_n_secs should be provided.
        3. Exactly one of log_dir and metric_logger should be provided.
    """
    super(LoggingMetricHook, self).__init__(
        tensors=tensors,
        every_n_iter=every_n_iter,
        every_n_secs=every_n_secs,
        at_end=at_end)

    if metric_logger is None:
      raise ValueError("metric_logger should be provided.")
    self._logger = metric_logger

  def begin(self):
    super(LoggingMetricHook, self).begin()
    self._global_step_tensor = tf.compat.v1.train.get_global_step()
    if self._global_step_tensor is None:
      raise RuntimeError(
          "Global step should be created to use LoggingMetricHook.")
    if self._global_step_tensor.name not in self._current_tensors:
      self._current_tensors[self._global_step_tensor.name] = (
          self._global_step_tensor)

  def after_run(self, unused_run_context, run_values):
    # should_trigger is a internal state that populated at before_run, and it is
    # using self_timer to determine whether it should trigger.
    if self._should_trigger:
      self._log_metric(run_values.results)

    self._iter_count += 1

  def end(self, session):
    if self._log_at_end:
      values = session.run(self._current_tensors)
      self._log_metric(values)

  def _log_metric(self, tensor_values):
    self._timer.update_last_triggered_step(self._iter_count)
    global_step = tensor_values[self._global_step_tensor.name]
    # self._tag_order is populated during the init of LoggingTensorHook
    for tag in self._tag_order:
      self._logger.log_metric(tag, tensor_values[tag], global_step=global_step)



# -*- coding: utf-8 -*-
from abc import ABCMeta, abstractmethod
from six import with_metaclass
from skydl.model.layers.old_layers import OldLayers


class SuperNet(with_metaclass(ABCMeta)):
    """
    由多个layer组成的网络
    usage:
    class MNISTTFNet(SuperTFNet):
        def __init__(self, name=None, parser_args=None):
            super().__init__(name, parser_args)
            self.conv1 = nn.Conv2d(1, 20, 5)
            self.conv2 = nn.Conv2d(20, 20, 5)

        def forward(self, x):
            super().forward(x)
            x = F.relu(self.conv1(x))
            return F.relu(self.conv2(x))
    """
    @property
    def parser_args(self):
        return self._parser_args

    def name(self):
        """
        子类可以重写该方法
        :return:
        # super().name()
        """
        return self._net_name if self._net_name else 'super_net'

    @property
    def model_proxy(self)->object:
        """
        model_proxy maybe None or keras.Model or torch.nn.Module
        子类可以重写该方法的返回值类型便于IDE找到对应的静态类. e.g.
        @property
        def model_proxy(self)->keras.Model:
            return self._model_proxy
        """
        return self._model_proxy

    def get_proxy(self):
        if self._model_proxy is None:
            return self
        return self._model_proxy

    def __init__(self, name=None, parser_args=None):
        """
        子类需要重写该方法
        :param name e.g. "mnist" 或 "gym_trading"
        # super().__init__(name, parser_args)
        """
        self._model_proxy = None
        self._net_name = name
        self._parser_args = parser_args

    @abstractmethod
    def forward(self, *input, **kwargs):
        """
        子类需要重写该方法
        目标函数，返回可以和target(label)对比的predict(logits)
        :param *input: tuple类型 input[0]->batch_x
        :param **kwargs: dict类型 kwargs.get("parser_args")->parser_args
        :return: result
        # super().forward(x)
        ......
        model = MNISTTFModel().build_network()
        result = model(batch_x)
        ......
        """
        # return self.call_hidden_layers(input[0], name=kwargs.get("name"))
        if len(input) < 1:  # for keras serving
            return self.call_hidden_layers(kwargs.get("inputs"), name="predict")
        return self.call_hidden_layers(input[0], name="predict")

    def __call__(self, *input, **kwargs):
        """
        model = MNISTTFModel().build_network()
        有了__call__方法就可以方便直接调用model(batch_x, parser_args=self.parser_args)方法
        来执行下面forward方法的内容
        :param input:
        :param kwargs:
        :return:
        """
        result = self.forward(*input, **kwargs)
        return result

    @abstractmethod
    def call_hidden_layers(self, _x, name="predict"):
        """
        子类可以重写该方法
        @:param name: predict节点名，便于model serving
        @see forward(input)
        """
        old_layers = OldLayers(self.parser_args)
        name = 'predict' if name is None else name
        if self.parser_args.model == 'cnn':
            predict = old_layers.cnn2_layers(_x, name=name)
        elif self.parser_args.model == 'keras_cnn':
            predict = old_layers.keras_cnn_layers(_x, name=name)
        elif self.parser_args.model == 'lstm':
            predict = old_layers.lstm_layers(_x, name=name)
        elif self.parser_args.model == 'bilstm':
            self.parser_args.num_hidden = 2
            predict = old_layers.bilstm_layers(_x, name=name)
        else:
            predict = None
        return predict

    # @autograph.convert()
    def huber_loss_just_for_test(a, tf_devices=None):
        import tensorflow as tf
        with tf_devices:
            if tf.abs(a) <= tf.constant(3.0):
                loss = a * a / 2.0
            else:
                loss = tf.constant(3.0) * (tf.abs(a) - tf.constant(3.0) / 2.0)
            return loss

    # @autograph.convert()
    @abstractmethod
    def compile_loss_metrics_optimizer_predict(self, _x, _y, _learning_rate=0.001, devices=[]):
        """
        子类可以重写该方法
        Define loss(cost) and optimizer
        ref: The art of regularization  https://greydanus.github.io/2016/09/05/regularization/
        https://github.com/greydanus/regularization
        TensorFlow with multiple GPUs https://jhui.github.io/2017/03/07/TensorFlow-GPU/
        [TensorFlow笔记] TensorArray解析 https://blog.csdn.net/guolindonggld/article/details/79256018
        AutoGraph converts Python into TensorFlow graphs  https://medium.com/tensorflow/autograph-converts-python-into-tensorflow-graphs-b2a871f87ec7
        @FIXME 后续将该函数纳入@autograph.convert()，以使多GPU的train的模型在serving后的Inference阶段即使没有GPU的机器上也能正确执行模型graph
        如果accuracy为None, 则可以返回default_accuracy = tf.constant(0)
        :param _x:
        :param _y:
        :param _learning_rate:
        :param devices: e.g. []
        :return:
        """
        import tensorflow as tf
        num_gpu = len(devices)
        losses = []
        correct_pred_total = []
        _predict = None
        if num_gpu > 1:
            '''
            @FIXME 注意该处需要修改成训练后的模型可以适配在任何设备上都能serving model。目前self.parser_args.use_cuda=True训练的模型在serving后的Inference会报错
            tf.split在batch_size=1时会出错
            先不显式控制GPU，让tensorflow自己控制GPU的数据处理
            '''
            # 单机多cpu训练
            X_A = tf.split(_x, num_gpu)
            Y_A = tf.split(_y, num_gpu)
            for gpu_id in range(num_gpu):
                with devices[gpu_id]:
                    with tf.variable_scope(tf.get_variable_scope(), reuse=tf.AUTO_REUSE):
                        _predict = self.forward(X_A[gpu_id])
                        target = Y_A[gpu_id]
                        cost = tf.nn.softmax_cross_entropy_with_logits_v2(logits=_predict, labels=target)
                        losses.append(cost)
                        # Evaluate model
                        correct_pred = tf.equal(tf.argmax(_predict, 1), tf.argmax(target, 1))
                        correct_pred_total.append(correct_pred)
            # losses = tf.convert_to_tensor(losses, dtype=tf.float32)
            loss = tf.reduce_mean(tf.concat(losses, axis=0))
            loss = tf.add_n([loss] + tf.get_collection(tf.GraphKeys.REGULARIZATION_LOSSES))
            tf.summary.scalar('loss', loss)
            # Evaluate model
            accuracy = tf.reduce_mean(tf.cast(tf.concat(correct_pred_total, axis=0), tf.float32))
            tf.summary.scalar('accuracy', accuracy)
        else:
            _predict = self.forward(_x)
            target = _y
            loss = tf.reduce_mean(tf.nn.softmax_cross_entropy_with_logits_v2(logits=_predict, labels=target))
            loss = tf.add_n([loss] + tf.get_collection(tf.GraphKeys.REGULARIZATION_LOSSES))
            tf.summary.scalar('loss', loss)
            correct_pred = tf.equal(tf.argmax(_predict, 1), tf.argmax(target, 1))
            accuracy = tf.reduce_mean(tf.cast(correct_pred, tf.float32))
            tf.summary.scalar('accuracy', accuracy)
        # #Important!, colocate_gradients_with_ops收集多卡GPU训练数据
        # optimizer = tf.train.AdamOptimizer(learning_rate=_learning_rate).minimize(loss, colocate_gradients_with_ops=True)
        # #用了下面这个optimizer后训练时间比上面的optimizer从23秒增加到了39秒
        import tensorflow.contrib.slim as slim
        optimizer = slim.learning.create_train_op(loss,
                                                  tf.train.AdamOptimizer(learning_rate=_learning_rate),
                                                  summarize_gradients=True,
                                                  colocate_gradients_with_ops=True)
        return loss, accuracy, optimizer, _predict




# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Central location for shared argparse convention definitions."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys
import codecs
import functools

from absl import app as absl_app
from absl import flags


# This codifies help string conventions and makes it easy to update them if
# necessary. Currently the only major effect is that help bodies start on the
# line after flags are listed. All flag definitions should wrap the text bodies
# with help wrap when calling DEFINE_*.
_help_wrap = functools.partial(flags.text_wrap, length=80, indent="",
                               firstline_indent="\n")


# Pretty formatting causes issues when utf-8 is not installed on a system.
def _stdout_utf8():
  try:
    codecs.lookup("utf-8")
  except LookupError:
    return False
  return sys.stdout.encoding == "UTF-8"


if _stdout_utf8():
  help_wrap = _help_wrap
else:
  def help_wrap(text, *args, **kwargs):
    return _help_wrap(text, *args, **kwargs).replace(u"\ufeff", u"")


# Replace None with h to also allow -h
absl_app.HelpshortFlag.SHORT_NAME = "h"



import argparse
import logging
import time
import ray
from ray.streaming import StreamingContext
from ray.streaming.config import Config

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

parser = argparse.ArgumentParser()
parser.add_argument(
    "--titles-file",
    required=True,
    help="the file containing the wikipedia titles to lookup")

# Splits input line into words and
# outputs records of the form (word,1)
def splitter(line):
    return [(word, 1) for word in line.split()]


if __name__ == "__main__":
    # Get program parameters
    args = parser.parse_args()
    titles_file = str(args.titles_file)
    ######################################
    # import ray
    # import time
    # ray.init(address="auto")
    # @ray.remote
    # def f(i):
    #     print("remote function....")
    #     time.sleep(10)
    #     return i
    # futures = [f.remote(i) for i in range(20)]
    # print(ray.get(futures))
    # ray.shutdown()
    import ray
    import time
    ray.init()
    @ray.remote
    def f():
        return b""
    n = 1000
    f_direct = f.options(is_direct_call=True)
    while True:
        start = time.time()
        ray.get([f_direct.remote() for _ in range(n)])
        print(n / (time.time() - start), "calls per second")
    ######################################
    # ray.init(address="auto", load_code_from_local=True, include_java=True)
    # ctx = StreamingContext.Builder() \
    #     .option(Config.CHANNEL_TYPE, Config.NATIVE_CHANNEL) \
    #     .build()
    # # A Ray streaming environment with the default configuration
    # ctx.set_parallelism(1)  # Each operator will be executed by two actors
    #
    # # Reads articles from wikipedia, splits them in words,
    # # shuffles words, and counts the occurrences of each word.
    # # stream = ctx.source(Wikipedia(titles_file)) \
    # # stream = ctx.source(ctx.read_text_file(titles_file)) \
    # stream = ctx.read_text_file(titles_file) \
    #     .flat_map(splitter) \
    #     .key_by(lambda x: x[0]) \
    #     .reduce(lambda old_value, new_value:
    #             (old_value[0], old_value[1] + new_value[1])) \
    #     .sink(print)
    # start = time.time()
    # ctx.execute("wordcount")
    # end = time.time()
    # logger.info("Elapsed time: {} secs".format(end - start))


# coding=utf-8
import sys
import re
import io
import unicodedata
import tensorflow as tf
import tensorflow_datasets as tfds
from tensorflow_datasets.core.features import text_feature
from py_common_util.common.annotations import Override
from skydl.datasets.super_tf_dataset_builder import DefaultDatasetBuilderConfig, SuperDatasetBuilder


class TranslateDatasetBuilder(SuperDatasetBuilder):
    """
    参考：https://www.tensorflow.org/tutorials/text/nmt_with_attention
    https://github.com/tensorflow/docs/blob/master/site/zh-cn/tutorials/text/nmt_with_attention.ipynb
    https://github.com/tensorflow/datasets/blob/master/tensorflow_datasets/translate/ted_hrlr.py
    https://rubikscode.net/2019/08/05/transformer-with-python-and-tensorflow-2-0-attention-layers/
    我们将使用http://www.manythings.org/anki/ 提供的一个语言数据集。这个数据集包含如下格式的语言翻译对：
    May I borrow this book? ¿Puedo tomar prestado este libro?
    """
    BUILDER_CONFIGS = [
        DefaultDatasetBuilderConfig(
            name="plain_text",
            version="0.0.1",
            description="Plain text",
        )
    ]
    _VALID_LANGUAGE_PAIRS = (
        ("en", "cn")
    )

    def __init__(self, data_dir=None, config=None):
        super().__init__(data_dir=data_dir, config=config)
        self._num_shards = 1
        self._train_data_file = sys.path[0] + "/data/translate/cmn-eng/cmn.txt"
        self._test_data_file = None
        self._validation_data_file = None

    @Override(SuperDatasetBuilder)
    def _info(self):
        return tfds.core.DatasetInfo(
            builder=self,
            description=("Chinese (Mandarin) - English dataset"),
            features=tfds.features.FeaturesDict({
                lang: text_feature.Text(None, None)
                for lang in self._VALID_LANGUAGE_PAIRS
            }),
            supervised_keys=self._VALID_LANGUAGE_PAIRS,
            urls=["http://www.manythings.org/anki/cmn-eng.zip"],
            citation="@ONLINE {Chinese (Mandarin) - English dataset}"
        )

    @Override(SuperDatasetBuilder)
    def _split_generators(self, dl_manager):
        generator_array = []
        if self._train_data_file:
            generator_array.append(tfds.core.SplitGenerator(
                name=tfds.Split.TRAIN,
                num_shards=3,
                gen_kwargs={
                    "data_path": self._train_data_file
                }))
        if self._test_data_file:
            generator_array.append(tfds.core.SplitGenerator(
                name=tfds.Split.TEST,
                num_shards=1,
                gen_kwargs={
                    "data_path": self._test_data_file
                }))
        if self._validation_data_file:
            generator_array.append(tfds.core.SplitGenerator(
                name=tfds.Split.VALIDATION,
                num_shards=1,
                gen_kwargs={
                    "data_path": self._validation_data_file
                }))
        return generator_array

    @Override(SuperDatasetBuilder)
    def _generate_examples(self, **kwargs):
        """Generate features and target given the directory path.
        Args:
          file_path: path where the csv file is stored
        Yields:
          The features and the target
        """
        # 这里已经手工下载好了zip文件，其实还可以在线下载文件
        # path_to_zip = tf.keras.utils.get_file('cmn-eng.zip', origin='http://www.manythings.org/anki/cmn-eng.zip', extract=True)
        # path_to_file = os.path.dirname(path_to_zip) + "/cmn-eng/cmn.txt"
        data_path = kwargs.get("data_path")
        with tf.io.gfile.GFile(data_path) as f:
            # input_tensor, target_tensor, inp_lang, targ_lang = self.load_dataset(f.name)
            # num_examples = input_tensor.shape[0]
            # for i in range(num_examples):
            #     yield {
            #         "en": target_tensor[i],
            #         "cn": input_tensor[i]
            #     }
            targ_lang, inp_lang = TranslateDatasetBuilder.create_dataset(f.name)
            for idx, (targ, inp) in enumerate(zip(targ_lang, inp_lang)):
                yield {
                    "en": targ,
                    "cn": inp
                }

    def unicode_to_ascii(s):
        """将 unicode 文件转换为 ascii"""
        return ''.join(c for c in unicodedata.normalize('NFD', s) if unicodedata.category(c) != 'Mn')

    def preprocess_sentence(w, is_chinese=False):
        """
        1. 去除重音符号 2. 清理句子 3. 返回这样格式的单词对：[ENGLISH, CHINESE]
        ```
        e.g.
        en_sentence = u"May I borrow this book?"
        sp_sentence = u"¿Puedo tomar prestado este libro?"
        print(preprocess_sentence(en_sentence))
        print(preprocess_sentence(sp_sentence).encode('utf-8'))
        ```
        :param w:
        :return:
        """
        w = TranslateDatasetBuilder.unicode_to_ascii(w.lower().strip())
        # 在单词与跟在其后的标点符号之间插入一个空格
        # 例如： "he is a boy." => "he is a boy ."
        # 参考：https://stackoverflow.com/questions/3645931/python-padding-punctuation-with-white-spaces-keeping-punctuation
        w = re.sub(r"([？?。.！!，,¿])", r" \1 ", w)
        w = re.sub(r'[" "]+', " ", w)
        # 除了中文汉字与(a-z, A-Z, ".", "?", "!", ",")，将其它所有字符替换为空格
        w = re.sub(r"[^a-zA-Z?.!,¿\u4e00-\u9fa5]+", " ", w)
        w = w.rstrip().strip()
        # 给句子加上开始和结束标记
        # 以便模型知道何时开始和结束预测
        if is_chinese:
            # 把中文句子拆成一个个的字 e.g. '把中文句子拆成一个个的字' -> '把 中 文 句 子 拆 成 一 个 个 的 字'
            w = ' '.join([t for t in w])
        w = '<start> ' + w + ' <end>'
        return w

    def create_dataset(path):
        """
        ```
        cmn.txt文件格式：
        Hi.	嗨。	CC-BY 2.0 (France) Attribution: tatoeba.org #538123 (CM) & #891077 (Martha)
        Hi.	你好。	CC-BY 2.0 (France) Attribution: tatoeba.org #538123 (CM) & #4857568 (musclegirlxyp)
        ```
        :param path:
        :return:
        """
        lines = io.open(path, encoding='UTF-8').read().strip().split('\n')
        word_pairs = [[TranslateDatasetBuilder.preprocess_sentence(w, False if idx < 1 else True) for idx, w in enumerate(l.split('\t')) if idx < 2] for l in lines]
        return zip(*word_pairs)

    def tokenize(lang):
        lang_tokenizer = tf.keras.preprocessing.text.Tokenizer(filters='')
        lang_tokenizer.fit_on_texts(lang)
        tensor = lang_tokenizer.texts_to_sequences(lang)
        tensor = tf.keras.preprocessing.sequence.pad_sequences(tensor, padding='post')
        return tensor, lang_tokenizer

    def load_dataset(path):
        # 创建清理过的输入输出对
        targ_lang, inp_lang = TranslateDatasetBuilder.create_dataset(path)
        return TranslateDatasetBuilder.load_dataset_from_sentences(targ_lang, inp_lang)

    def load_dataset_from_sentences(targ_lang, inp_lang):
        """
        将全局的word生成对应的int tokenize
        :param targ_lang: e.g. ["<start> i m dead tired . <end>", "<start> don t forget smoking is bad for your health . <end>"]
        :param inp_lang:  e.g. ["<start> 我 累 死 了 <end>", "<start> 不 要 忘 记 吸 烟 对 你 的 健 康 有 害 <end>"]
        :return:
        """
        input_tensor, inp_lang_tokenizer = TranslateDatasetBuilder.tokenize(inp_lang)
        target_tensor, targ_lang_tokenizer = TranslateDatasetBuilder.tokenize(targ_lang)
        return input_tensor, target_tensor, inp_lang_tokenizer, targ_lang_tokenizer

    def max_length(tensor):
        """计算目标张量的最大长度（max_length）"""
        return max(len(t) for t in tensor)

    def convert(lang, tensor):
        """
        ```
        返回 e.g.
        1 ----> <start>
        2605 ----> mande
        ```
        :param tensor:
        :return:
        """
        for t in tensor:
            if t != 0:
                print("%d ----> %s" % (t, lang.index_word[t]))


if __name__ == '__main__':
    dataset_name = TranslateDatasetBuilder.camelcase_to_snakecase("TranslateDatasetBuilder")
    print(SuperDatasetBuilder.list_builders())
    # num_examples = 21621
    num_examples = TranslateDatasetBuilder.get_num_examples(dataset_name, tfds.Split.TRAIN)
    batched_datasets = TranslateDatasetBuilder.load_batched_datasets(dataset_name,
                                                                     config_name="plain_text",
                                                                     split=[tfds.Split.TRAIN],
                                                                     batch_size=num_examples,
                                                                     shuffle=-1)
    batched_dataset_features = TranslateDatasetBuilder.read_batched_datasets(batched_datasets)
    all_cn_sentences = []
    all_en_sentences = []
    for features, labels in batched_dataset_features:
        all_en_sentences += [feature.decode('utf-8') for feature in features]
        all_cn_sentences += [label.decode('utf-8') for label in labels]
        # print(f"features={''.join([feature.decode('utf-8') for feature in features])}, labels={''.join([label.decode('utf-8') for label in labels])}")
    input_tensor, target_tensor, inp_lang, targ_lang = TranslateDatasetBuilder.load_dataset_from_sentences(all_en_sentences, all_cn_sentences)
    # 计算目标张量的最大长度 （max_length）
    max_length_targ, max_length_inp = TranslateDatasetBuilder.max_length(target_tensor), TranslateDatasetBuilder.max_length(input_tensor)
    print(f"max_length_targ={max_length_targ}, max_length_inp={max_length_inp}")
    TranslateDatasetBuilder.convert(inp_lang, input_tensor[0])
    print(f"end!!")
    #dataset = tf.data.Dataset.from_tensor_slices(np.array([1.0, 2.0, 3.0, 4.0, 5.0]))
    dataset = tf.data.Dataset.from_tensor_slices((input_tensor, target_tensor)).batch(64, drop_remainder=True)
    # for input_batch, target_batch in tfds.as_numpy(dataset):
    #     print(input_batch, target_batch)




# -*- coding: utf-8 -*-
import numpy as np
import tensorflow_datasets as tfds
from skydl.models.train_phase_enum import TrainPhaseEnum
from skydl.datasets.movielens_dataset_builder import MovieLensDatasetBuilder
from skydl.models_zoo.cv.minist_tf_keras_modelv2_main import training_mnist_tf_keras_model
from skydl.examples.modelsv2.playground_models.dummy_tf_keras_modelv2 import DummyTfKerasModelV2
from skydl.models_zoo.recommendation.wide_and_deep_tf_keras_modelv2_main import training_wide_and_deep_keras_model


def run_example_0():
    def load_data_fn(parser_args):
        print(f"load_data_fn......{parser_args}")
        dataset_name = MovieLensDatasetBuilder.camelcase_to_snakecase("MovieLensDatasetBuilder")
        num_examples = MovieLensDatasetBuilder.get_num_examples(dataset_name)
        full_datasets = MovieLensDatasetBuilder.load_batched_datasets(dataset_name,
                                                                      split=[tfds.Split.TRAIN],
                                                                      batch_size=128,
                                                                      epochs=1)
        train_data, validation_data, evaluate_data = MovieLensDatasetBuilder.split_full_datasets(full_datasets,
                                                                                                      num_examples=MovieLensDatasetBuilder.get_total_num_batch(
                                                                                                          num_examples,
                                                                                                          128),
                                                                                                      train_size=0.7,
                                                                                                      validation_size=0,
                                                                                                      test_size=0.2)
        np_user_ids = np.array([])
        np_item_ids = np.array([])
        np_labels = np.array([])
        if parser_args.train_phase == TrainPhaseEnum.Train.value:
            # get 2 batched test_dataset: i.e. 128 * 3
            stat_dataset = evaluate_data.take(3)
        else:
            # get train_dataset
            stat_dataset = train_data
        for features, labels in MovieLensDatasetBuilder.read_batched_datasets(stat_dataset):
            np_user_ids = np.append(np_user_ids, features['user_id'])
            np_item_ids = np.append(np_item_ids, features['item_id'])
            np_labels = np.append(np_labels, labels)
        return train_data, validation_data, evaluate_data, np_user_ids, np_item_ids, np_labels
    # kick of training
    DummyTfKerasModelV2("TfKerasModelV2-0",
    ).set_load_data_fn(
        load_data_fn
    ).compile(
    ).fit(
    )


def run_example_1():
    training_wide_and_deep_keras_model()


def run_mnist():
    training_mnist_tf_keras_model()


if __name__ == '__main__':
    # run_example_0()
    # run_example_1()
    run_mnist()




# -*- coding: utf-8 -*-
import sys
import os
import tensorflow as tf
from skydl.models_zoo.bert.official.nlp.bert_models import classifier_model
from skydl.models_zoo.bert.official.nlp.bert.tokenization import FullTokenizer, printable_text, convert_to_unicode
from transformers import BertConfig
import tensorflow_hub as hub


class InputExample(object):
    """A single training/test example for simple sequence classification."""

    def __init__(self, guid, text, label=None):
        """Constructs a InputExample.

        Args:
          guid: Unique id for the example.
          text_a: string. The untokenized text of the first sequence. For single
            sequence tasks, only this sequence must be specified.
          label: (Optional) string. The label of the example. This should be
            specified for train and dev examples, but not for test examples.
        """
        self.guid = guid
        self.text = text
        self.label = label


class InputFeatures(object):
    """A single set of features of data."""

    def __init__(self, input_ids, input_mask, segment_ids, label_ids,):
        self.input_ids = input_ids
        self.input_mask = input_mask
        self.segment_ids = segment_ids
        self.label_ids = label_ids
        #self.label_mask = label_mask


class DataProcessor(object):
    """Base class for data converters for sequence classification data sets."""

    def get_train_examples(self, data_dir):
        """Gets a collection of `InputExample`s for the train set."""
        raise NotImplementedError()

    def get_dev_examples(self, data_dir):
        """Gets a collection of `InputExample`s for the dev set."""
        raise NotImplementedError()

    def get_labels(self):
        """Gets the list of labels for this data set."""
        raise NotImplementedError()

    @classmethod
    def _read_data(cls, input_file):
        """Reads a BIO data."""
        with open(input_file) as f:
            lines = []
            words = []
            labels = []
            for line in f:
                contends = line.strip()
                word = line.strip().split(' ')[0]
                label = line.strip().split(' ')[-1]
                if contends.startswith("-DOCSTART-"):
                    words.append('')
                    continue
                # if len(contends) == 0 and words[-1] == '。':
                if len(contends) == 0:
                    l = ' '.join([label for label in labels if len(label) > 0])
                    w = ' '.join([word for word in words if len(word) > 0])
                    lines.append([l, w])
                    words = []
                    labels = []
                    continue
                words.append(word)
                labels.append(label)
            return lines


class NerProcessor(DataProcessor):
    def get_train_examples(self, data_dir):
        return self._create_example(
            self._read_data(os.path.join(data_dir, "train.txt")), "train"
        )

    def get_dev_examples(self, data_dir):
        return self._create_example(
            self._read_data(os.path.join(data_dir, "dev.txt")), "dev"
        )

    def get_test_examples(self,data_dir):
        return self._create_example(
            self._read_data(os.path.join(data_dir, "test.txt")), "test")


    def get_labels(self):
        # prevent potential bug for chinese text mixed with english text
        # return ["O", "B-PER", "I-PER", "B-ORG", "I-ORG", "B-LOC", "I-LOC", "[CLS]","[SEP]"]
        return ["O", "B-PER", "I-PER", "B-ORG", "I-ORG", "B-LOC", "I-LOC", "X","[CLS]","[SEP]"]

    def _create_example(self, lines, set_type):
        examples = []
        for (i, line) in enumerate(lines):
            guid = "%s-%s" % (set_type, i)
            text = convert_to_unicode(line[1])
            label = convert_to_unicode(line[0])
            examples.append(InputExample(guid=guid, text=text, label=label))
        return examples


def create_bert_model(num_labels=1):
    """run ner model"""
    model_path = sys.path[0] + "/../../datasets/data/bert/tf_hub/bert_zh_L-12_H-768_A-12/1"
    # load config
    config = BertConfig.from_json_file(sys.path[0] + "/bert_config.json")
    MAX_SEQUENCE_LENGTH = config.max_position_embeddings
    HIDDEN_DROPOUT_PROB = config.hidden_dropout_prob
    INITIALIZER_RANGE = config.initializer_range
    USE_AMP = False
    # define inputs, outputs
    input_word_ids = tf.keras.layers.Input(shape=(MAX_SEQUENCE_LENGTH,), dtype=tf.int32, name="input_word_ids")
    input_masks = tf.keras.layers.Input(shape=(MAX_SEQUENCE_LENGTH,), dtype=tf.int32, name="input_masks")
    segment_ids = tf.keras.layers.Input(shape=(MAX_SEQUENCE_LENGTH,), dtype=tf.int32, name="segment_ids")
    # build model
    bert_model = hub.KerasLayer(model_path, trainable=True)
    # pooled_output用于分类，sequence_output用于序列相关的预测
    pooled_output, sequence_output = bert_model([input_word_ids, input_masks, segment_ids])
    sequence_output.layers[-1].activation = tf.keras.activations.softmax
    output = tf.keras.layers.Dropout(rate=HIDDEN_DROPOUT_PROB)(pooled_output)
    output = tf.keras.layers.Dense(num_labels,
                                   kernel_initializer=tf.keras.initializers.TruncatedNormal(stddev=INITIALIZER_RANGE),
                                   name='output',
                                   dtype=tf.float32)(output)
    # build tokenizer
    vocab_file = bert_model.resolved_object.vocab_file.asset_path.numpy()
    do_lower_case = bert_model.resolved_object.do_lower_case.numpy()
    tokenizer = FullTokenizer(vocab_file, do_lower_case)
    # build classifier model
    classifier_model = tf.keras.Model(
        inputs=[input_word_ids, input_masks, segment_ids],
        outputs=output)
    # compile model
    metric = tf.keras.metrics.SparseCategoricalAccuracy("accuracy")
    if num_labels == 1:
        loss = tf.keras.losses.MeanSquaredError()
    else:
        loss = tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True)
    # Prepare training: Compile tf.keras model with optimizer, loss and learning rate schedule
    optimizer = tf.keras.optimizers.Adam(learning_rate=3e-5, epsilon=1e-08, clipnorm=1.0)
    if USE_AMP:
        # loss scaling is currently required when using mixed precision
        optimizer = tf.keras.mixed_precision.experimental.LossScaleOptimizer(optimizer, "dynamic")
    classifier_model.compile(optimizer=optimizer, loss=loss, metrics=[metric])
    classifier_model.summary()
    return classifier_model, bert_model


def convert_single_example(ex_index, example, label_map, max_seq_length, tokenizer):
    textlist = example.text.split(' ')
    labellist = example.label.split(' ')
    tokens = []
    labels = []
    # print(textlist)
    for i, word in enumerate(textlist):
        token = tokenizer.tokenize(word)
        # print(token)
        tokens.extend(token)
        label_1 = labellist[i]
        # print(label_1)
        for m in range(len(token)):
            if m == 0:
                labels.append(label_1)
            else:
                labels.append("X")
        # print(tokens, labels)
    # tokens = tokenizer.tokenize(example.text)
    if len(tokens) >= max_seq_length - 1:
        tokens = tokens[0:(max_seq_length - 2)]
        labels = labels[0:(max_seq_length - 2)]
    ntokens = []
    segment_ids = []
    label_ids = []
    ntokens.append("[CLS]")
    segment_ids.append(0)
    # append("O") or append("[CLS]") not sure!
    label_ids.append(label_map["[CLS]"])
    for i, token in enumerate(tokens):
        ntokens.append(token)
        segment_ids.append(0)
        label_ids.append(label_map[labels[i]])
    ntokens.append("[SEP]")
    segment_ids.append(0)
    # append("O") or append("[SEP]") not sure!
    label_ids.append(label_map["[SEP]"])
    input_ids = tokenizer.convert_tokens_to_ids(ntokens)
    input_mask = [1] * len(input_ids)
    #label_mask = [1] * len(input_ids)
    while len(input_ids) < max_seq_length:
        input_ids.append(0)
        input_mask.append(0)
        segment_ids.append(0)
        # we don't concerned about it!
        label_ids.append(0)
        ntokens.append("**NULL**")
        #label_mask.append(0)
    # print(len(input_ids))
    assert len(input_ids) == max_seq_length
    assert len(input_mask) == max_seq_length
    assert len(segment_ids) == max_seq_length
    assert len(label_ids) == max_seq_length
    #assert len(label_mask) == max_seq_length

    if ex_index < 5:
        print("*** Example ***")
        print("guid: %s" % (example.guid))
        print("tokens: %s" % " ".join([printable_text(x) for x in tokens]))
        print("input_ids: %s" % " ".join([str(x) for x in input_ids]))
        print("input_mask: %s" % " ".join([str(x) for x in input_mask]))
        print("segment_ids: %s" % " ".join([str(x) for x in segment_ids]))
        print("label_ids: %s" % " ".join([str(x) for x in label_ids]))
        #print("label_mask: %s" % " ".join([str(x) for x in label_mask]))

    feature = InputFeatures(
        input_ids=input_ids,
        input_mask=input_mask,
        segment_ids=segment_ids,
        label_ids=label_ids,
        #label_mask = label_mask
    )
    return feature


def process_data(max_seq_length=128):
    model_path = sys.path[0] + "/../../datasets/data/bert/tf_hub/bert_zh_L-12_H-768_A-12/1"
    data_dir = sys.path[0] + "/data/ner"
    processor = NerProcessor()
    label_list = processor.get_labels()
    num_labels = len(label_list) + 1
    tokenizer = FullTokenizer(vocab_file=model_path + "/assets/vocab.txt", do_lower_case=True)  # todo
    examples = processor.get_dev_examples(data_dir)
    label_map = {}
    for (i, label) in enumerate(label_list, 1):
        label_map[label] = i
    for (ex_index, example) in enumerate(examples):
        if ex_index % 5000 == 0:
            print("Writing example %d of %d" % (ex_index, len(examples)))
        feature = convert_single_example(ex_index, example, label_map, max_seq_length, tokenizer)


if __name__ == '__main__':
    # bert_model, tokenizer = create_bert_model(num_labels=3)
    process_data(max_seq_length=128)




# -*- coding: utf-8 -*-
from abc import ABCMeta, abstractmethod
import tensorflow as tf
from six import with_metaclass
import numpy as np
from py_common_util.common.annotations import print_exec_time
from py_common_util.tensorflow.tf_utils import TFUtils
import argparse
import datetime, time
import os,sys
from logbook import Logger, StreamHandler, FileHandler
from collections import OrderedDict
from skydl.model.constants import Constants
from skydl.model.super_net import SuperNet
from skydl.model.train_phase_enum import TrainPhaseEnum
sys.path.append(os.path.abspath('%s/..' % sys.path[0]))


class SuperModel(with_metaclass(ABCMeta)):
    """
    tf keras model
    class WideDeepKerasMode(DefaultKerasModel):
    定义model class：
    class XxxKerasModel(DefaultKerasModel):
        def adjust_parse_args_value(self):
            super().adjust_parse_args_value()
        def load_data(self, *args, **kwargs):
            super().load_data(*args, **kwargs)
        def fit(self, *args, **kwargs):
            return self
        def evaluate(self, *args, **kwargs):
            return self
        def serving(self, *args, **kwargs):
            return self
        def predict(self, *args, **kwargs):
            if not self.is_inference_phase():
                return None
            return None

    调用例子如下：
    from tensorflow.python import keras
    MyKerasModel("my_keras_model", [
        keras.layers.Flatten(input_shape=(28, 28)),
        keras.layers.Dropout(0.2),
        keras.layers.Dense(128, activation='relu'),
        keras.layers.Dense(10, activation='softmax', name="prediction")
    ]).compile(
        keras.losses.sparse_categorical_crossentropy,
        keras.optimizers.Adam(),
        ['accuracy']
    ).fit().serving()
    ==========
    参考列表：
    https://www.tensorflow.org/alpha
    https://www.tensorflow.org/alpha/guide/saved_model#serving_the_model
    Training and Serving ML models with tf.keras  https://medium.com/tensorflow/training-and-serving-ml-models-with-tf-keras-fd975cc0fa27
    https://github.com/ray-project/ray/blob/master/python/ray/experimental/sgd/model.py
    https://colab.research.google.com
    将您的代码升级至 TensorFlow 2.0 https://mp.weixin.qq.com/s/BD-nJSZJLjBBq1n7HEHpKw
    tf2例子：https://github.com/tensorflow/docs/tree/master/site/en/r2/guide
    我的tf2训练场： https://colab.research.google.com/drive/1LtRP9lSivkdXJHX3S1_O8lFa9s21U6-V#scrollTo=ev01UBIHvaOj
    TensorFlow 2.0 Alpha 版发布啦！赶紧来尝鲜！https://mp.weixin.qq.com/s/gwcVOYhlEEBv28QoaFSnbw
    手工升级v1的代码到v2：$tf_upgrade_v2 --infile /Users/tony/myfiles/spark/share/python-projects/deep_trading/neural_networks/model/super_tf_model.py --outfile /Users/tony/myfiles/spark/share/python-projects/deep_trading/neural_networks/model/super_tf_model2.py
    """
    @property
    def log(self):
        return self._log

    @log.setter
    def log(self, value):
        self._log = value

    @property
    def parser(self):
        return self._parser

    @property
    def parser_args(self):
        return self._parser_args

    @property
    def num_gpu(self):
        return self._num_gpu

    @property
    def device(self):
        """
        gpu设备数组，多个设备表示为[tf.device(),tf.device()]
        为充分利用gpu, batch_size必须设置尽量大
        :return:
        usage:
        .........
        with self.device:
            sum_operation = tf.reduce_sum(dot_operation)
            ......
        """
        return self._device

    @property
    def name(self):
        return self._name

    @property
    def model(self)->SuperNet:
        """
        定义SuperNet返回值类型便于IDE找到对应的静态类
        子类可以重写该方法的返回值类型便于IDE找到对应的静态类. e.g.
        @property
        def model(self)->estimator.Estimator:
            return self._model
        """
        return self._model

    @model.setter
    def model(self, value):
        self._model = value

    @property
    def loss(self):
        """Return loss of the model
        Returns:
            loss
        """
        return self._loss

    @loss.setter
    def loss(self, value):
        self._loss = value

    @property
    def optimizer(self):
        """Return optimizer for the model
        Returns:
            optimizer
        """
        return self._optimizer

    @optimizer.setter
    def optimizer(self, value):
        self._optimizer = value

    @property
    def metrics(self):
        """Return metrics of the model
        Returns:
            metrics(dict): e.g. {"accuracy": accuracy(numpy data)}
        """
        return self._metrics

    @metrics.setter
    def metrics(self, value):
        self._metrics = value

    @property
    def weights(self):
        """Return weights from the model.
        Implementing `get_weights` is required for checkpointing and fault
        tolerance.
        Returns:
            Numpy array of weights from the model.
        """
        return self._weights

    @weights.setter
    def set_weights(self, value):
        self._weights = value

    @property
    def latest_model_filename(self):
        return "model.ckpt"

    @property
    def distribute_strategy(self):
        """
        多GPU分布式训练策略
        e.g. print('Number of devices: {}'.format(self.distribute_strategy.num_replicas_in_sync))
        """
        return self._distribute_strategy

    def __init__(self, name=None, loss=None, optimizer=None, metrics=None, weights=None):
        super().__init__()
        StreamHandler(sys.stdout).push_application()
        self._name = name
        self._log = Logger(self.name if self.name else self.__class__.__name__)
        self._parser = argparse.ArgumentParser()
        self._parser_args = None
        self._device = None
        self._num_gpu = 0
        self.do_parse_args()
        self._model = None if self.parser_args.lazy_load_model else self.build_network()
        # loss, optimizer, metrics, weights
        self._loss = loss
        self._optimizer = optimizer
        self._metrics = metrics
        self._weights = weights
        # distribute strategy, e.g. tf.distribute.MirroredStrategy(devices=self.device)
        self._distribute_strategy = tf.distribute.MirroredStrategy()

    @abstractmethod
    def build_network(self):
        """
        子类需要重写该方法
        :return: a net of SuperTfNets
        # super().build_network()
        """
        return SuperNet(self.name if self.name else 'super_net', self.parser_args)

    def get_model_checkpoint_dir(self):
        model_checkpoint_dir = self.parser_args.saved_model_path + "/" + self.model.name() + "/" + self.parser_args.model_version
        if not os.path.exists(model_checkpoint_dir):
            os.makedirs(model_checkpoint_dir)
        return model_checkpoint_dir

    def get_model_file_saved_dir(self):
        """
        整个model文件保存的目录：
        for keras, 使用如：
        ```
        # save mode to h5
        model.save(self.get_model_file_saved_dir() + "/saved_model.h5")
        # restore model from h5
        model = tf.keras.models.load_model(self.get_model_file_saved_dir() + "/saved_model.h5")
        ```
        :return:
        """
        model_file_saved_dir = "/tmp/super_model_file_saved_dir/" + self.model.name() + "/" + self.parser_args.model_version
        if not os.path.exists(model_file_saved_dir):
            os.makedirs(model_file_saved_dir)
        return model_file_saved_dir


    @abstractmethod
    def load_data(self):
        """
        load train|valid|test dataset
        子类需要重写该方法
        :return: lazy_load_data, iter_train_data_labels, total_train_batch, train data, test data, valid data, num_train_data, num_test_data, num_valid_data, input_shape_channel_height_width
        # super().load_data()
        """
        dict_load_data = OrderedDict()
        dict_load_data[Constants.DICT_LOAD_DATA_LAZY_LOAD_DATA] = False,
        dict_load_data[Constants.DICT_LOAD_DATA_ITER_TRAIN_DATA_LABELS] = tf.compat.v1.data.make_one_shot_iterator(tf.data.Dataset.range(1)),
        dict_load_data[Constants.DICT_LOAD_DATA_TOTAL_TRAIN_BATCH] = 0,
        dict_load_data[Constants.DICT_LOAD_DATA_NUM_TRAIN_DATA] = [],
        dict_load_data[Constants.DICT_LOAD_DATA_TEST_DATA] = [],
        dict_load_data[Constants.DICT_LOAD_DATA_VALIDATION_DATA] = [],
        dict_load_data[Constants.DICT_LOAD_DATA_NUM_TRAIN_DATA] = 0,
        dict_load_data[Constants.DICT_LOAD_DATA_NUM_TEST_DATA] = 0,
        dict_load_data[Constants.DICT_LOAD_DATA_NUM_VALIDATION_DATA] = 0,
        dict_load_data[Constants.DICT_LOAD_DATA_SHAPE_CHANNEL_HEIGHT_WIDTH] = (1, 28, 28)
        return dict_load_data

    def do_define_x_y(self):
        """子类可以重写该方法"""
        if self.parser_args.channels > 1:
            _x = tf.compat.v1.placeholder("float32", [None, self.parser_args.time_steps, self.parser_args.rnn_input, self.parser_args.channels])
        else:
            _x = tf.compat.v1.placeholder("float32", [None, self.parser_args.time_steps, self.parser_args.rnn_input])
        _y = tf.compat.v1.placeholder("int64", [None, self.parser_args.classes])
        return _x, _y

    def add_parser_argument(self):
        """
        增加parser参数项
        子类需要重写该方法
        # super().add_parser_argument()
        """
        self.parser.add_argument('--init_from_saver', type=bool, default=True, metavar='N',
                            help='init from saved checkpoint')
        self.parser.add_argument('--saved_model_path', type=str, default=sys.path[0] + '/saved_model',
                            help='model checkpoint saved to path')
        self.parser.add_argument('--model_version', type=str, default='1',
                                 help='version number of the model')
        self.parser.add_argument('--train_phase', type=str, default=TrainPhaseEnum.Train.value,
                            help='train->train phase, test->test phase, inference->online inference phase, freeze->freeze model')
        self.parser.add_argument('--tf_summary_enabled', type=bool, default=False, metavar='N',
                            help='是否开启tensorflow的summary功能，注意在单机多GPU模式下开启该功能会让训练时间从24秒增加到107秒')
        self.parser.add_argument('--summary_path', type=str, default=sys.path[0] + '/summary',
                            help='dir where TensorBoard summary save to')
        self.parser.add_argument('--decay_rate', type=float, default=0.97, metavar='N',
                            help='decay rate for rmsprop, Decaying the learning rate)')
        self.parser.add_argument('--model', type=str, default='lstm',
                            help='value of "cnn|lstm|bilstm", if bilstm then should set--num-hidden=2 else --num-hidden=1')
        self.parser.add_argument('--channels', type=int, default=1, metavar='N',
                            help='channels (i.e. depth, default: 1)')
        self.parser.add_argument('--height', type=int, default=28,
                            help='height')
        self.parser.add_argument('--width', type=int, default=28,
                            help='width')
        self.parser.add_argument('--num_classes', type=int, default=10,
                            help='num of classes')
        self.parser.add_argument('--time_steps', type=int, default=28,
                            help="time steps num of RNN, default equals to height's value, also equals to seq_length")
        self.parser.add_argument('--rnn_input', type=int, default=28,
                            help="RNN input(the value as same as width)")
        self.parser.add_argument('--rnn_hidden', type=int, default=256,
                            help="num of RNN hidden_layer or state size, default value is multiple of batch_size's value")
        self.parser.add_argument('--num_hidden', type=int, default=1,
                            help="num of n_hidden, for RNN, default value is 1: using rnn.static_bidirectional_rnn")
        self.parser.add_argument('--keep_prob', type=float, default=0.5,
                                 help="dropout: keep probability, default value is: 0.5")
        # 以下变量和pytorch共用
        self.parser.add_argument('--data_path', type=str, default=sys.path[0] + '/dataset',
                            help='location of the loading data corpus')
        self.parser.add_argument('--onnx_export_path', type=str, default=sys.path[0] + '/saved_model',
                                 help='path to export the final model in onnx format')
        self.parser.add_argument('--batch_size', type=int, default=64, metavar='N',
                            help='input batch size for training (default: 64)')
        self.parser.add_argument('--eval_batch_size', type=int, default=1000, metavar='N',
                            help='input batch size for testing (default: 1000)')
        self.parser.add_argument('--epochs', type=int, default=1, metavar='N',
                            help='number of epochs to train (default: 10)')
        self.parser.add_argument('--learning_rate', type=float, default=0.001, metavar='LR',
                            help='learning rate (default: 0.001)')
        self.parser.add_argument('--momentum', type=float, default=0.5, metavar='M',
                            help='SGD momentum (default: 0.5)')
        self.parser.add_argument('--use_cuda', action='store_true', default=True,
                            help='use CUDA')
        self.parser.add_argument('--seed', type=int, default=1, metavar='S',
                            help='random seed (default: 1)')
        self.parser.add_argument('--log_interval', type=int, default=1000, metavar='N',
                            help='how many batches to wait before logging training status')
        self.parser.add_argument('--lazy_load_model', type=bool, default=False, metavar='N',
                             help='lazy load model(build_network())')

    def adjust_parse_args_value(self):
        """
        调整pare_args参数的值
        子类需要重写该方法
        修改parser_args的值
        # super().adjust_parse_args_value()
        e.g. self.parser_args.seed = 1
        参考代码：
        def adjust_parse_args_value(self):
            super().adjust_parse_args_value()
            self.parser_args.data_path = sys.path[0] + '/../../dataset'
            self.parser_args.use_cuda = True
            self.parser_args.init_from_saver = True
            self.parser_args.train_phase = TrainPhaseEnum.Train.value
            self.parser_args.model_version = '1'
            self.parser_args.epochs = 1
            self.parser_args.batch_size = 128
            self.parser_args.log_interval = 1000
        """
        pass

    def do_parse_args(self):
        self.add_parser_argument()
        self._parser_args = self.parser.parse_args()
        self.adjust_parse_args_value()
        # 设置random seed确保每次训练都可以获得相同唯一的的随机序列
        np.random.seed(self.parser_args.seed)
        tf.compat.v1.set_random_seed(self.parser_args.seed)
        self._num_gpu = TFUtils.check_available_gpus(show_info=True)
        if self.num_gpu > 0:
            if not self.parser_args.use_cuda:
                self.log.info("WARNING: You have " + str(self.num_gpu) + " CUDA device, so you should probably run with --cuda")
                # 不使用GPU加速
                self._num_gpu = 0
        else:
            self.parser_args.use_cuda = False
        if self.parser_args.use_cuda:
            self._device = [TFUtils.device("gpu" if self.parser_args.use_cuda else "cpu", gpu_id) for gpu_id in range(self._num_gpu)]
        else:
            self._device = []
        self.log.info("Tensorflow Version: " + tf.__version__ + "\nMain function argparse: " + str(self.parser_args))

    def is_training_phase(self):
        return self.parser_args.train_phase == TrainPhaseEnum.Train.value

    def is_evaluation_phase(self):
        return self.parser_args.train_phase == TrainPhaseEnum.Test.value

    def is_serving_phase(self):
        return self.parser_args.train_phase == TrainPhaseEnum.Serving.value

    def is_inference_phase(self):
        return self.parser_args.train_phase == TrainPhaseEnum.Inference.value

    def compile(self, loss=None, optimizer=None, metrics=None):
        """
        Configures the model for training
        compile a model before training/testing
        inference阶段不需要调用模型的compile函数
        @see keras.models.Model.compile()
        :param loss:
        :param optimizer:
        :param metrics:
        :return: self
        """
        if self.parser_args.lazy_load_model:
            self.model = self.build_network()
        return self

    @abstractmethod
    def fit(self, *args, **kwargs):
        """
        模型训练的启动入口
        子类需要重写该方法
        Training settings: 设置一些参数，每个都有默认值，输入 $python3 main.py -h 可以获得相关帮助
        $python3 main.py -batch_size=32 -log_interval=20
        :return: self
        """
        # kwargs = {'num_workers': 1, 'pin_memory': True} if self.parser_args.use_cuda else {}
        if self.parser_args.train_phase == TrainPhaseEnum.Freeze.value:
            # handle freeze model phase
            self.freeze_graph()
            return
        elif self.parser_args.train_phase == TrainPhaseEnum.Inference.value:
            return self.inference()
        # load train|valid|test dataset
        lazy_load_data, \
        iter_train_data_labels, \
        total_train_batch, \
        train_data, \
        valid_data, \
        test_data, \
        num_train_total, \
        num_valid_total, \
        num_test_total, \
        shape_channels_height_width = self.load_data()
        self.parser_args.channels, self.parser_args.height, self.parser_args.width = shape_channels_height_width
        self.log.info(self.model.name() + '==>>> lazy load data mode: {}'.format(lazy_load_data))
        self.log.info(self.model.name() + '==>>> total training batch number: {}'.format(total_train_batch))
        self.log.info(self.model.name() + '==>>> total validation batch number: {}'.format(len(valid_data)))
        self.log.info(self.model.name() + '==>>> total testing batch number: {}'.format(len(test_data)))
        self.log.info('[' + self.model.name() + ']begin to fitting model, Time: {}'.format(datetime.datetime.now()))
        with tf.compat.v1.name_scope('do_tf_training'):
            # add iter_train_data_labels to saver
            iter_batch_train_data_labels = iter_train_data_labels.get_next()
            saveable_obj_iterator = tf.data.experimental.make_saveable_from_iterator(iter_train_data_labels)
            tf.compat.v1.add_to_collection(tf.compat.v1.GraphKeys.SAVEABLE_OBJECTS, saveable_obj_iterator)
            # add learning_rate|epoch|step to saver
            train_learning_rate = tf.Variable(initial_value=0.0, trainable=False, dtype=tf.float32)
            curr_epoch_of_num_epoch = tf.Variable(initial_value=-1, trainable=False, dtype=tf.int32)
            curr_step_of_total_batch = tf.Variable(initial_value=-1, trainable=False, dtype=tf.int32)
            _x, _y = self.do_define_x_y()
            ###############
            # from tensorflow.contrib import autograph
            # tf_huber_loss = autograph.to_code(self.model.huber_loss_just_for_test)
            # print("tf_huber_loss, code==", tf_huber_loss)
            # # tf_huber_loss = autograph.to_graph(self.model.huber_loss_just_for_test)
            # # print("tf_huber_loss==", tf.Session().run(tf_huber_loss(tf.constant(12.0), self.device[0])))
            # # tf_do_loss_accuracy_optimizer_predict = autograph.to_code(self.model.do_loss_accuracy_optimizer_predict)
            # tf_do_loss_accuracy_optimizer_predict = autograph.to_code(self.model.do_loss_accuracy_optimizer_predict)
            # print("tf_do_loss_accuracy_optimizer_predict, code=", tf_do_loss_accuracy_optimizer_predict)
            # # print(autograph.to_code(self.model.do_loss_accuracy_optimizer_predict))
            ###############
            cost, accuracy, optimizer, predict = self.model.compile_loss_metrics_optimizer_predict(_x, _y, train_learning_rate, self.device)
            # allow_soft_placement=True: 如果指定的GPU设备不存在，允许TF自动分配CPU设备
            sess = tf.compat.v1.Session(config=tf.compat.v1.ConfigProto(allow_soft_placement=True, log_device_placement=False))
            init_op = tf.group(tf.compat.v1.global_variables_initializer(), tf.compat.v1.local_variables_initializer())  # sess.run(tf.tables_initializer())
            sess.run(init_op)
            saver = tf.compat.v1.train.Saver()
            try:
                start_time = time.time()
                if self.parser_args.train_phase == TrainPhaseEnum.Train.value:
                    # init_from_saver: True-从saved_model对应项目的缓存中拿上次的model结果作为起始参数，也可以手工删除saved_model对应项目的缓存文件以重新训练
                    if self.parser_args.init_from_saver:
                        checkpoint = tf.train.get_checkpoint_state(
                            self.get_model_checkpoint_dir(),
                            latest_filename=self.latest_model_filename)
                        if checkpoint and checkpoint.saved_model_path:
                            saver.restore(sess, checkpoint.saved_model_path)
                            self.log.info("Restored and Init Session Variables from Saver......")
                        else:
                            self.log.info("*** Error Occurred in " + self.parser_args.train_phase + " phase(1), Can not find the model file: "
                                + self.get_model_checkpoint_dir() + "/" + self.latest_model_filename)
                            self.log.info("And you can ignore this error, then the training will go on......")
                    self.log.info("TensorBoard Summary will Writing to {}".format(self.parser_args.summary_path))
                    self.log.info("Run the command line:\n" \
                          "--> tensorboard --logdir=" + self.parser_args.summary_path + "/../.." \
                          + "\nThen open http://0.0.0.0:6006/ into your web browser")
                    if self.parser_args.tf_summary_enabled:
                        tf_summary_writer = tf.compat.v1.summary.FileWriter(self.parser_args.summary_path, sess.graph)
                        merged_summary_op = tf.compat.v1.summary.merge_all()
                    from_step = sess.run(curr_step_of_total_batch) + 1
                    from_epoch = sess.run(curr_epoch_of_num_epoch)
                    self.log.info("init start: from_epoch_index=%d, from_step=%d" % (from_epoch, from_step))
                    if from_epoch < 0:
                        from_epoch += 1
                    else:
                        if from_step >= (total_train_batch - 1):
                            from_epoch += 1
                            from_step = 0
                    self.log.info("init end: from_epoch_index=%d, from_step=%d" % (from_epoch, from_step))
                    epoch_runnable = False
                    for epoch in range(from_epoch, self.parser_args.epochs):
                        self.log.info("begin to training data(%d), display_step: %d, epoch/(total): %d/%d ......"
                              % (num_train_total, self.parser_args.log_interval, epoch + 1, self.parser_args.epochs))
                        sess.run(curr_epoch_of_num_epoch.assign(epoch))
                        learning_rate = sess.run(tf.compat.v1.assign(train_learning_rate, self.parser_args.learning_rate * (self.parser_args.decay_rate ** epoch)))
                        total_cost = 0
                        if epoch > from_epoch:
                            from_step = 0
                        for step in range(from_step, total_train_batch):
                            epoch_runnable = True
                            batch_x, batch_y = sess.run(iter_batch_train_data_labels) if lazy_load_data else train_data[step]
                            batch_x, batch_y = self.do_transform_training_batched_data(batch_x, batch_y)
                            if not self.parser_args.tf_summary_enabled:
                                _, loss, acc = sess.run([optimizer, cost, accuracy],
                                                             feed_dict={_x: batch_x,
                                                                        _y: batch_y,
                                                                        train_learning_rate: learning_rate})
                            else:
                                merged_summarys, _, loss, acc = sess.run([merged_summary_op, optimizer, cost, accuracy],
                                                                         feed_dict={_x: batch_x,
                                                                                    _y: batch_y,
                                                                                    train_learning_rate: learning_rate})
                                tf_summary_writer.add_summary(merged_summarys, step)
                            total_cost += loss
                            if ((step + 1) % self.parser_args.log_interval == 0) or ((step + 1) >= total_train_batch):
                                sess.run(curr_step_of_total_batch.assign(step))
                                saver.save(sess,
                                           self.get_model_checkpoint_dir() + "/model.ckpt",
                                           global_step=epoch*total_train_batch+step,
                                           latest_filename="model.ckpt")
                                self.log.info("epoch+1/(total)/step+1/(total)/iter_data/(total): "
                                      + str(epoch + 1) + "/" + str(self.parser_args.epochs)
                                      + "/" + str(step + 1) + "/" + str(total_train_batch) + "/"
                                      + str((step + 1) * self.parser_args.batch_size) + "/" + str(num_train_total)
                                      + ", total_loss=" + "{:.6f}".format(total_cost)
                                      + ", training_accuracy=" + "{:.2f}%".format(100.00 * acc))
                    # save session after all training steps
                    if epoch_runnable:
                        saver.save(sess,
                                   self.get_model_checkpoint_dir() + "/model.ckpt",
                                   global_step=epoch*total_train_batch+step,
                                   latest_filename="model.ckpt")
                        tf.io.write_graph(sess.graph.as_graph_def(from_version=1),
                                             self.get_model_checkpoint_dir(),
                                             'graph.pbtxt',
                                             as_text=True)
                        self.log.info("saved model file to: " + self.parser_args.saved_model_path + "/" + self.model.name())
                        # 统计整个训练时间
                        duration = (time.time() - start_time) * 1000
                        self.log.info("training Duration(milliseconds)=%d, Optimization Finished!" % duration)
                checkpoint = tf.train.get_checkpoint_state(
                    self.get_model_checkpoint_dir(),
                    latest_filename=self.latest_model_filename)
                if checkpoint and checkpoint.saved_model_path:
                    saver.restore(sess, checkpoint.saved_model_path)
                else:
                    self.log.info("*** Error For " + self.parser_args.train_phase
                          + " phase(2), can not find the model file: "
                          + self.get_model_checkpoint_dir() + "/" + self.latest_model_filename)
                    return
                self.log.info("Success to restore the model file for %s phase!" % self.parser_args.train_phase)
                if self.parser_args.train_phase == TrainPhaseEnum.Train.value or self.parser_args.train_phase == TrainPhaseEnum.Test.value:
                    self.log.info("Begin to Calculating Accuracy for all Test data(%d)......" % num_test_total)
                    start_time = time.time()
                    acc = self.do_test_from_numpy_data(sess, _x, _y, accuracy, test_data)
                    self.log.info("Done Test and the Test Accuracy is {:.2f}%".format(100.0*acc))
                    duration = (time.time() - start_time) * 1000
                    self.log.info(self.model.name() + ": for validate of " + self.parser_args.train_phase + " phase, Duration(milliseconds)=%d, Validate Finished!" % duration)
                    self.log.info("Now:" + str(datetime.datetime.now()))
                if self.parser_args.train_phase == TrainPhaseEnum.Train.value or self.parser_args.train_phase == TrainPhaseEnum.Serving.value:
                    self.build_serving_model(sess,
                                             _x, predict,
                                             self.get_model_checkpoint_dir())
            except:
                self.log.info('Error Occurred at trainning ' + self.model.name() + '......')
                import traceback
                traceback.print_exc()
            finally:
                pass
            return self

    @abstractmethod
    def evaluate(self, *args, **kwargs):
        """
        在测试阶段测试训练好的模型效果
        @:param evaluate_dataset:测试数据集, shape: [num_batched, batch_size, feature_size,label_size]
        :return: self
        ```
        usage:
         evaluate(evaluate_dataset=[]) then: evaluate_dataset = kwargs.get("evaluate_dataset")
         evaluate([]) then: evaluate_dataset = args[0])
        ```
        """
        return self

    @abstractmethod
    def serving(self):
        """
        build serving model
        用docker serving开启服务
        client可以使用java直接调用模型
        =====================
        https://www.tensorflow.org/alpha/guide/saved_model#serving_the_model
        java请求tensorflow-serving服务，返回模型结果  http://www.zhongruitech.com/222460444.html
        https://github.com/tobegit3hub/tensorflow_template_application/tree/master/java_predict_client
        使用Java客户端  http://docs.api.xiaomi.com/cloud-ml/modelservice/0903_use_java_client.html
        https://stackoverflow.com/questions/47196792/how-to-use-java-client-to-request-tensorflow-serving-for-widedeep-model-or-how
        构建 TensorFlow Serving Java 客户端  https://kaiyuanshuwu.com/articles/138
        =====================
        启动serving服务
        $docker run -p 8500:8500 -p 8501:8501 \
        --mount type=bind,source=/Users/tony/myfiles/spark/share/python-projects/deep_trading/skydl/examples/saved_model/serving/models/my_keras_model,target=/models/my_keras_model \
        -e MODEL_NAME=my_keras_model -t tensorflow/serving:1.13.0
        :return: self
        """
        return self

    @abstractmethod
    def predict(self, *args, **kwargs):
        """python版的预测接口，返回预测结果"""
        return None

    @print_exec_time
    def inference(self, tfx_host_port="http://localhost:8501/", model_name=None, features_list=[]):
        """
        inference request to tf serving
        https://medium.com/tensorflow/serving-ml-quickly-with-tensorflow-serving-and-docker-7df7094aa008
        Deploy TensorFlow models https://towardsdatascience.com/deploy-tensorflow-models-9813b5a705d5
        #########################################
        $docker rmi -f <<old serving images>>
        $docker pull tensorflow/serving:1.13.0
        $docker run -p 8500:8500 -p 8501:8501 \
        --mount type=bind,source=/Users/tony/myfiles/spark/share/python-projects/deep_trading/skydl/examples/saved_model/serving/models/default_keras_net,target=/models/my_keras_model \
        -e MODEL_NAME=my_keras_model -t tensorflow/serving:1.13.0
        ########################################
        1.RESTful API https://www.tensorflow.org/tfx/serving/api_rest#model_metadata_api
        http://192.168.83.111:8501/v1/models/user_book_recommend/versions/1/metadata 或者 http://192.168.83.111:8501/v1/models/user_book_recommend/metadata
        2.A Java Client for TensorFlow Serving gRPC API https://medium.com/@junwan01/a-java-client-for-tensorflow-serving-grpc-api-d37b5ad747aa
        3.Tensorflow: save model for use in Java or C++  https://medium.com/@alexkn15/tensorflow-save-model-for-use-in-java-or-c-ab351a708ee4
        4.Saving and Loading a TensorFlow model using the SavedModel API  https://medium.com/@jsflo.dev/saving-and-loading-a-tensorflow-model-using-the-savedmodel-api-17645576527
        #########################################
        :param tfx_host_port: docker serving url, e.g. "http://localhost:8501/" 完整的post url，如"http://localhost:8501/v1/models/my_keras_model:predict"
        :param model_name: inference_model.model.name() e.g. "my_keras_model"
        :param features_list:  shape e.g. (1,28,28), 数据格式：tf_dataset_batched_features.tolist()[:100]
        features_list数据可以来自如下：
        _, evaluate_dataset = inference_model.load_data()
        import tensorflow_datasets as tfds
        for dataset in tfds.as_numpy(evaluate_dataset):
            features, labels = dataset
        features_list = features.tolist()
        ###用Postman工具调用的例子：
        #POST: http://192.168.83.111:8501/v1/models/user_book_recommend:predict
        Headers: {"content-type": "application/json"}
        Body(json):
        {
            "signature_name": "serving_default",
            "instances": [{"user_id":[0], "item_id":[260]}, {"user_id":[1], "item_id":[260]}]
        }
        Response:
        ```
        {
            "predictions": [
                [
                    1.50607
                ],
                [
                    3.53324
                ]
            ]
        }
        ```
        #GET: http://192.168.83.111:8501/v1/models/user_book_recommend/metadata
        Response:
        ```
        {
            "model_spec": {
                "name": "user_book_recommend",
                "signature_name": "",
                "version": "9"
            },
            "metadata": {
                "signature_def": {
                    "signature_def": {
                        "serving_default": {
                            "inputs": {
                                "user_id": {
                                    "dtype": "DT_FLOAT",
                                    "tensor_shape": {
                                        "dim": [
                                            {
                                                "size": "-1",
                                                "name": ""
                                            },
                                            {
                                                "size": "1",
                                                "name": ""
                                            }
                                        ],
                                        "unknown_rank": false
                                    },
                                    "name": "serving_default_user_id:0"
                                },
                                "item_id": {
                                    "dtype": "DT_FLOAT",
                                    "tensor_shape": {
                                        "dim": [
                                            {
                                                "size": "-1",
                                                "name": ""
                                            },
                                            {
                                                "size": "1",
                                                "name": ""
                                            }
                                        ],
                                        "unknown_rank": false
                                    },
                                    "name": "serving_default_item_id:0"
                                }
                            },
                            "outputs": {
                                "dense": {
                                    "dtype": "DT_FLOAT",
                                    "tensor_shape": {
                                        "dim": [
                                            {
                                                "size": "-1",
                                                "name": ""
                                            },
                                            {
                                                "size": "1",
                                                "name": ""
                                            }
                                        ],
                                        "unknown_rank": false
                                    },
                                    "name": "StatefulPartitionedCall:0"
                                }
                            },
                            "method_name": "tensorflow/serving/predict"
                        },
                        "__saved_model_init_op": {
                            "inputs": {},
                            "outputs": {
                                "__saved_model_init_op": {
                                    "dtype": "DT_INVALID",
                                    "tensor_shape": {
                                        "dim": [],
                                        "unknown_rank": true
                                    },
                                    "name": "NoOp"
                                }
                            },
                            "method_name": ""
                        }
                    }
                }
            }
        }
        ```
        :return:
        """
        import json
        import requests
        import traceback
        data = json.dumps({"signature_name": "serving_default",
                           "instances": features_list})
        headers = {"content-type": "application/json"}
        # 去掉url里最后一个/字符
        if tfx_host_port[-1] == '/':
            tfx_host_port = tfx_host_port[:-1]
        post_url = tfx_host_port + "/v1/models/" + model_name + ":predict"
        json_response = requests.post(post_url, data=data, headers=headers)
        print("the inference request takes times(milliseconds): " + str(json_response.elapsed.total_seconds()*1000))
        try:
            if not json_response.ok:
                self.log.info("the inference request failure, url: " + json_response.url + ", response code: " + str(json_response.status_code) + ", error reason: " + json_response.text)
                return []
            print("inference response json predictions：", json_response.json()["predictions"])
            print("inference response json predictions(删除掉单维度)：", np.squeeze(json_response.json()["predictions"]))
            prediction_list = np.array(json_response.json()["predictions"])
            print("inference response is ok, prediction_list after argmax is:" + str(tf.argmax(prediction_list, axis=1).numpy()))
            return prediction_list
        except:
            print("exception occured at an inference request!")
            traceback.print_exc()
        return []

    def build_serving_model(self,
                            sess=None,
                            _x=None,
                            predict=None,
                            saved_model_path=None,
                            data_record_name=None,
                            model_version='1'):
        """
        TODO 该方法将被废弃
        #Serving Inception Model with TensorFlow Serving and Kubernetes https://www.tensorflow.org/serving/serving_inception
        #Improve TensorFlow Serving Performance With GPU Support https://docs.bitnami.com/google/how-to/enable-nvidia-gpu-tensorflow-serving/
        #tensorflow将训练好的模型freeze,即将权重固化到图里面,并使用该模型进行预测 https://blog.csdn.net/lujiandong1/article/details/53385092
        # A Tool Developer's Guide to TensorFlow Model Files https://www.tensorflow.org/extend/tool_developers/
        # How to EASILY put Machine Learning Models into Production using Tensorflow Serving
        https://medium.com/coinmonks/how-to-easily-put-machine-learning-models-into-production-using-tensorflow-serving-91998fa4b4e1
        # https://github.com/nex3z/tfserving-mnist
        # ref: https://www.tensorflow.org/serving/serving_basic
        # ref: https://github.com/tensorflow/serving/blob/master/tensorflow_serving/example/mnist_client.py
        https://hub.docker.com/r/tensorflow/serving/
        tensorflow教程6：Supervisor长期训练帮手  https://www.jianshu.com/p/7490ebfa3de8
        #########################################
        $docker rmi -f <<old serving images>>
        $docker pull tensorflow/serving:1.13.0
        $docker run -p 8500:8500 -p 8501:8501 \
        --mount type=bind,source=/tony/pycharm_projects/deep_trading/saved_model/serving/models/kcws,target=/models/kcws \
        -e MODEL_NAME=kcws -t tensorflow/serving:1.13.0
        """
        if saved_model_path is None:
            saved_model_path = self.parser_args.saved_model_path + '/serving/models'
        if data_record_name is None:
            data_record_name = self.model.name()
        export_path_base = saved_model_path + '/' + data_record_name
        export_path = os.path.join(
            tf.compat.as_bytes(export_path_base),
            tf.compat.as_bytes(model_version))
        self.log.info('exporting trained model to', str(export_path, 'utf-8'))
        if os.path.exists(export_path):
            self.log.info('......serving model saved path is already exists and please retry again，'
                  'you can remove the model path manually, command: rm -rf ' + str(export_path, 'utf-8'))
            return
        builder = tf.compat.v1.saved_model.builder.SavedModelBuilder(export_path)
        tensor_info_x = tf.compat.v1.saved_model.utils.build_tensor_info(_x)
        tensor_info_y = tf.compat.v1.saved_model.utils.build_tensor_info(predict)
        prediction_signature = (
            tf.compat.v1.saved_model.signature_def_utils.build_signature_def(
                inputs={'inputs': tensor_info_x},
                outputs={'scores': tensor_info_y},
                method_name=tf.saved_model.PREDICT_METHOD_NAME))
        """
        设置strip_default_attrs=Ture会保存graph里面的tf.data.one_shot_iterator()对象, 会导致启动serving的docker时出现如下异常：
        Invalid argument: Cannot add function 'tf_data_structured_function_wrapper_IcY4sd7IC5Q' because a different function with the same name already exists.
        """
        builder.add_meta_graph_and_variables(
            sess, [tf.saved_model.SERVING],
            signature_def_map={
                'prediction': prediction_signature
            },
            main_op=tf.compat.v1.tables_initializer(),
            strip_default_attrs=False)
        builder.save()
        self.log.info('done exporting serving model: ' + export_path_base)

    def export_to_onnx(self, model, input_shape_channel_height_width):
        """
        TODO 该方法将被废弃
        :param model 继承SuperTfNet(nn.Module)
        :param input_channel_height_width_shape 为model(input)的input的shape e.g. 【1, 28, 28】
        :return:
        """
        pass

    def freeze_graph(self, input_graph=None,
                     input_saver='',
                     input_binary=False,
                     input_checkpoint=None,
                     output_node_names='do_tf_training/LF/predict/MatMul',
                     restore_op_name='do_tf_training/save/restore_all',
                     filename_tensor_name='do_tf_training/save/Const:0',
                     output_graph=None,
                     clear_devices=True,
                     initializer_nodes='',
                     variable_names_blacklist=''):
        """
        TODO 该方法将被废弃
        子类可以重写该方法
        Converts all variables in a graph and checkpoint into constants
        参考：https://github.com/tensorflow/tensorflow/blob/master/tensorflow/python/tools/freeze_graph.py
        :param input_graph: TensorFlow \'GraphDef\' file to load.
        :param input_saver: TensorFlow saver file to load.
        :param input_binary: Whether the input files are in binary format.
        :param input_checkpoint: TensorFlow variables file to load.
        :param output_node_names: The name of the output nodes, comma separated.
        :param restore_op_name: The name of the master restore operator.
        :param filename_tensor_name: The name of the tensor holding the save path.
        :param output_graph: Output \'GraphDef\' file name.
        :param clear_devices: Whether to remove device specifications.
        :param initializer_nodes: comma separated list of initializer nodes to run before freezing.
        :param variable_names_blacklist: comma separated list of variables to skip converting to constants
        :return:
        """
        from google.protobuf import text_format
        from tensorflow.core.framework import graph_pb2
        from tensorflow.core.protobuf import saver_pb2
        from tensorflow.python.client import session
        from tensorflow.python.framework import graph_util
        from tensorflow.python.framework import importer
        from tensorflow.python.platform import gfile
        from tensorflow.python.training import saver as saver_lib

        if input_graph is None:
            input_graph = self.get_model_checkpoint_dir() + '/graph.pbtxt'
        if input_checkpoint is None:
            checkpoint = tf.train.get_checkpoint_state(self.get_model_checkpoint_dir(),
                                                       latest_filename="model.ckpt")
            if checkpoint:
                input_checkpoint = checkpoint.saved_model_path
        if output_graph is None:
            output_graph = self.get_model_checkpoint_dir() + '/freezed-model.pbtxt'
        if not gfile.Exists(input_graph):
            self.log.info("Input graph file '" + input_graph + "' does not exist!")
            return -1

        if input_saver and not gfile.Exists(input_saver):
            self.log.info("Input saver file '" + input_saver + "' does not exist!")
            return -1

        # 'input_checkpoint' may be a prefix if we're using Saver V2 format
        if not saver_lib.checkpoint_exists(input_checkpoint):
            self.log.info("Input checkpoint '" + input_checkpoint + "' doesn't exist!")
            return -1

        if not output_node_names:
            self.log.info("You need to supply the name of a node to --output_node_names.")
            return -1

        input_graph_def = graph_pb2.GraphDef()
        mode = "rb" if input_binary else "r"
        with gfile.FastGFile(input_graph, mode) as f:
            if input_binary:
                input_graph_def.ParseFromString(f.read())
            else:
                # FIXED text_format.Merge(f.read().decode("utf-8"), input_graph_def)
                text_format.Merge(f.read(), input_graph_def)
        # Remove all the explicit device specifications for this node. This helps to
        # make the graph more portable.
        if clear_devices:
            for node in input_graph_def.node:
                node.device = ""
        _ = importer.import_graph_def(input_graph_def, name="")

        with session.Session() as sess:
            if input_saver:
                with gfile.FastGFile(input_saver, mode) as f:
                    saver_def = saver_pb2.SaverDef()
                    if input_binary:
                        saver_def.ParseFromString(f.read())
                    else:
                        text_format.Merge(f.read(), saver_def)
                    saver = saver_lib.Saver(saver_def=saver_def)
                    saver.restore(sess, input_checkpoint)
            else:
                sess.run([restore_op_name], {filename_tensor_name: input_checkpoint})
                if initializer_nodes:
                    sess.run(initializer_nodes)

            variable_names_blacklist = (variable_names_blacklist.split(",") if
                                        variable_names_blacklist else None)
            output_graph_def = graph_util.convert_variables_to_constants(
                sess,
                input_graph_def,
                output_node_names.split(","),
                variable_names_blacklist=variable_names_blacklist)
        with gfile.GFile(output_graph, "wb") as f:
            f.write(output_graph_def.SerializeToString())
        self.log.info("done freeze graph, %d ops in the final graph." % len(output_graph_def.node))

    def do_inference_from_freezed_model(self, freezed_model_path=None):
        """
        TODO 该方法将被废弃
        子类可以重写该方法
        :param freezed_model_path:
        :return:
        """
        return None

    def do_inference_from_saved_serving_model(self, np_x_input=None, model_name=None, hostport="localhost:8500"):
        """
        TODO 该方法将被废弃
        子类可以重写该方法
        :param np_x_input:
        :param model_name:
        :param hostport:
        :return:
        """
        import grpc
        from tensorflow_serving.apis import predict_pb2
        from tensorflow_serving.apis import prediction_service_pb2_grpc
        channel = grpc.insecure_channel(hostport)
        stub = prediction_service_pb2_grpc.PredictionServiceStub(channel)
        request = predict_pb2.PredictRequest()
        request.model_spec.name = model_name  # e.g. mnist
        request.model_spec.signature_name = 'serving_default'
        request.inputs['input_1'].CopyFrom(
            tf.compat.v1.make_tensor_proto(np_x_input, shape=np_x_input.shape))  # shape:e.g. [1,28,28,1]
        result_future = stub.Predict.future(request, 5.0)  # 5 seconds timeout
        exception = result_future.exception()
        if exception:
            self.log.info(exception)
        else:
            prediction = np.array(result_future.result().outputs['scores'].float_val)
            print("perdiction=", prediction)
            return prediction




# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Gaussian error linear unit."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math

import tensorflow as tf


@tf.keras.utils.register_keras_serializable(package='Text')
def gelu(x):
  """Gaussian Error Linear Unit.
  This is a smoother version of the RELU.
  Original paper: https://arxiv.org/abs/1606.08415
  Args:
    x: float Tensor to perform activation.
  Returns:
    `x` with the GELU activation applied.
  """
  cdf = 0.5 * (1.0 + tf.tanh(
      (math.sqrt(2 / math.pi) * (x + 0.044715 * tf.pow(x, 3)))))
  return x * cdf


# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Defining common flags used across all BERT models/applications."""

from absl import flags
import tensorflow as tf

from skydl.models_zoo.bert.official.utils.flags import core as flags_core


def define_common_bert_flags():
  """Define common flags for BERT tasks."""
  flags_core.define_base(
      data_dir=False,
      model_dir=True,
      clean=False,
      train_epochs=False,
      epochs_between_evals=False,
      stop_threshold=False,
      batch_size=False,
      num_gpu=True,
      hooks=False,
      export_dir=False,
      distribution_strategy=True,
      run_eagerly=True)
  flags.DEFINE_string('bert_config_file', None,
                      'Bert configuration file to define core bert layers.')
  flags.DEFINE_string(
      'model_export_path', None,
      'Path to the directory, where trainined model will be '
      'exported.')
  flags.DEFINE_string('tpu', '', 'TPU address to connect to.')
  flags.DEFINE_string(
      'init_checkpoint', None,
      'Initial checkpoint (usually from a pre-trained BERT model).')
  flags.DEFINE_integer('num_train_epochs', 3,
                       'Total number of training epochs to perform.')
  flags.DEFINE_integer(
      'steps_per_loop', 200,
      'Number of steps per graph-mode loop. Only training step '
      'happens inside the loop. Callbacks will not be called '
      'inside.')
  flags.DEFINE_float('learning_rate', 5e-5,
                     'The initial learning rate for Adam.')
  flags.DEFINE_boolean(
      'scale_loss', False,
      'Whether to divide the loss by number of replica inside the per-replica '
      'loss function.')
  flags.DEFINE_boolean(
      'use_keras_compile_fit', False,
      'If True, uses Keras compile/fit() API for training logic. Otherwise '
      'use custom training loop.')
  flags.DEFINE_string(
      'hub_module_url', None, 'TF-Hub path/url to Bert module. '
      'If specified, init_checkpoint flag should not be used.')
  flags.DEFINE_enum(
      'model_type', 'bert', ['bert', 'albert'],
      'Specifies the type of the model. '
      'If "bert", will use canonical BERT; if "albert", will use ALBERT model.')

  # Adds flags for mixed precision training.
  flags_core.define_performance(
      num_parallel_calls=False,
      inter_op=False,
      intra_op=False,
      synthetic_data=False,
      max_train_steps=False,
      dtype=True,
      dynamic_loss_scale=True,
      loss_scale=True,
      all_reduce_alg=False,
      num_packs=False,
      enable_xla=True,
      fp16_implementation=True,
  )


def use_float16():
  return flags_core.get_tf_dtype(flags.FLAGS) == tf.float16


def get_loss_scale():
  return flags_core.get_loss_scale(flags.FLAGS, default_for_fp16='dynamic')



from skydl.ray.streaming_0_8_2.python.streaming import Environment, ExecutionGraph
from skydl.ray.streaming_0_8_2.python.operator import OpType, PStrategy


def test_parallelism():
    """Tests operator parallelism."""
    env = Environment()
    # Try setting a common parallelism for all operators
    env.set_parallelism(2)
    stream = env.source(None).map(None).filter(None).flat_map(None)
    env._collect_garbage()
    for operator in env.operators.values():
        if operator.type == OpType.Source:
            # TODO (john): Currently each source has only one instance
            assert operator.num_instances == 1, (operator.num_instances, 1)
        else:
            assert operator.num_instances == 2, (operator.num_instances, 2)
    # Check again after adding an operator with different parallelism
    stream.map(None, "Map1").shuffle().set_parallelism(3).map(
        None, "Map2").set_parallelism(4)
    env._collect_garbage()
    for operator in env.operators.values():
        if operator.type == OpType.Source:
            assert operator.num_instances == 1, (operator.num_instances, 1)
        elif operator.name != "Map1" and operator.name != "Map2":
            assert operator.num_instances == 2, (operator.num_instances, 2)
        elif operator.name != "Map2":
            assert operator.num_instances == 3, (operator.num_instances, 3)
        else:
            assert operator.num_instances == 4, (operator.num_instances, 4)


def test_partitioning():
    """Tests stream partitioning."""
    env = Environment()
    # Try defining multiple partitioning strategies for the same stream
    _ = env.source(None).shuffle().rescale().broadcast().map(
        None).broadcast().shuffle()
    env._collect_garbage()
    for operator in env.operators.values():
        p_schemes = operator.partitioning_strategies
        for scheme in p_schemes.values():
            # Only last defined strategy should be kept
            if operator.type == OpType.Source:
                assert scheme.strategy == PStrategy.Broadcast, (
                    scheme.strategy, PStrategy.Broadcast)
            else:
                assert scheme.strategy == PStrategy.Shuffle, (
                    scheme.strategy, PStrategy.Shuffle)


def test_forking():
    """Tests stream forking."""
    env = Environment()
    # Try forking a stream
    stream = env.source(None).map(None).set_parallelism(2)
    # First branch with a shuffle partitioning strategy
    _ = stream.shuffle().key_by(0).sum(1)
    # Second branch with the default partitioning strategy
    _ = stream.key_by(1).sum(2)
    env._collect_garbage()
    # Operator ids
    source_id = None
    map_id = None
    keyby1_id = None
    keyby2_id = None
    sum1_id = None
    sum2_id = None
    # Collect ids
    for id, operator in env.operators.items():
        if operator.type == OpType.Source:
            source_id = id
        elif operator.type == OpType.Map:
            map_id = id
        elif operator.type == OpType.KeyBy:
            if operator.other_args == 0:
                keyby1_id = id
            else:
                assert operator.other_args == 1, (operator.other_args, 1)
                keyby2_id = id
        elif operator.type == OpType.Sum:
            if operator.other_args == 1:
                sum1_id = id
            else:
                assert operator.other_args == 2, (operator.other_args, 2)
                sum2_id = id
    # Check generated streams and their partitioning
    for source, destination in env.logical_topo.edges:
        operator = env.operators[source]
        if source == source_id:
            assert destination == map_id, (destination, map_id)
        elif source == map_id:
            p_scheme = operator.partitioning_strategies[destination]
            strategy = p_scheme.strategy
            key_index = env.operators[destination].other_args
            if key_index == 0:  # This must be the first branch
                assert strategy == PStrategy.Shuffle, (strategy,
                                                       PStrategy.Shuffle)
                assert destination == keyby1_id, (destination, keyby1_id)
            else:  # This must be the second branch
                assert key_index == 1, (key_index, 1)
                assert strategy == PStrategy.Forward, (strategy,
                                                       PStrategy.Forward)
                assert destination == keyby2_id, (destination, keyby2_id)
        elif source == keyby1_id or source == keyby2_id:
            p_scheme = operator.partitioning_strategies[destination]
            strategy = p_scheme.strategy
            key_index = env.operators[destination].other_args
            if key_index == 1:  # This must be the first branch
                assert strategy == PStrategy.ShuffleByKey, (
                    strategy, PStrategy.ShuffleByKey)
                assert destination == sum1_id, (destination, sum1_id)
            else:  # This must be the second branch
                assert key_index == 2, (key_index, 2)
                assert strategy == PStrategy.ShuffleByKey, (
                    strategy, PStrategy.ShuffleByKey)
                assert destination == sum2_id, (destination, sum2_id)
        else:  # This must be a sum operator
            assert operator.type == OpType.Sum, (operator.type, OpType.Sum)


def _test_shuffle_channels():
    """Tests shuffling connectivity."""
    env = Environment()
    # Try defining a shuffle
    _ = env.source(None).shuffle().map(None).set_parallelism(4)
    expected = [(0, 0), (0, 1), (0, 2), (0, 3)]
    _test_channels(env, expected)


def _test_forward_channels():
    """Tests forward connectivity."""
    env = Environment()
    # Try the default partitioning strategy
    _ = env.source(None).set_parallelism(4).map(None).set_parallelism(2)
    expected = [(0, 0), (1, 1), (2, 0), (3, 1)]
    _test_channels(env, expected)


def _test_broadcast_channels():
    """Tests broadcast connectivity."""
    env = Environment()
    # Try broadcasting
    _ = env.source(None).set_parallelism(4).broadcast().map(
        None).set_parallelism(2)
    expected = [(0, 0), (0, 1), (1, 0), (1, 1), (2, 0), (2, 1), (3, 0), (3, 1)]
    _test_channels(env, expected)


def _test_round_robin_channels():
    """Tests round-robin connectivity."""
    env = Environment()
    # Try broadcasting
    _ = env.source(None).round_robin().map(None).set_parallelism(2)
    expected = [(0, 0), (0, 1)]
    _test_channels(env, expected)


def _test_channels(environment, expected_channels):
    """Tests operator connectivity."""
    environment._collect_garbage()
    map_id = None
    # Get id
    for id, operator in environment.operators.items():
        if operator.type == OpType.Map:
            map_id = id
    # Collect channels
    environment.execution_graph = ExecutionGraph(environment)
    environment.execution_graph.build_channels()
    channels_per_destination = []
    for operator in environment.operators.values():
        channels_per_destination.append(
            environment.execution_graph._generate_channels(operator))
    # Check actual connectivity
    actual = []
    for destination in channels_per_destination:
        for channels in destination.values():
            for channel in channels:
                src_instance_index = channel.src_instance_index
                dst_instance_index = channel.dst_instance_index
                connection = (src_instance_index, dst_instance_index)
                assert channel.dst_operator_id == map_id, (
                    channel.dst_operator_id, map_id)
                actual.append(connection)
    # Make sure connections are as expected
    set_1 = set(expected_channels)
    set_2 = set(actual)
    assert set_1 == set_2, (set_1, set_2)


def test_channel_generation():
    """Tests data channel generation."""
    _test_shuffle_channels()
    _test_broadcast_channels()
    _test_round_robin_channels()
    _test_forward_channels()


# TODO (john): Add simple wordcount test
def test_wordcount():
    """Tests a simple streaming wordcount."""
    pass


if __name__ == "__main__":
    test_channel_generation()



# -*- coding: utf-8 -*-
import os
import time
import numpy as np
import tensorflow as tf
import tensorflow_datasets as tfds
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
from py_common_util.common.annotations import print_exec_time
from skydl.datasets.super_tf_dataset_builder import SuperDatasetBuilder
from skydl.datasets.translate_dataset_builder import TranslateDatasetBuilder
from skydl.models_zoo.attention.attention import BahdanauAttention
from skydl.models_zoo.attention.decoder import Decoder
from skydl.models_zoo.attention.encoder import Encoder

###############################
# global parameters
EPOCHS = 10
BATCH_SIZE = 64
embedding_dim = 256
units = 1024
###############################
dataset_name = TranslateDatasetBuilder.camelcase_to_snakecase("TranslateDatasetBuilder")
# num_examples = 21621
num_examples = TranslateDatasetBuilder.get_num_examples(dataset_name, tfds.Split.TRAIN)
batched_datasets = TranslateDatasetBuilder.load_batched_datasets(dataset_name,
                                                                 config_name="plain_text",
                                                                 split=[tfds.Split.TRAIN],
                                                                 batch_size=num_examples,
                                                                 shuffle=-1)
batched_dataset_features = TranslateDatasetBuilder.read_batched_datasets(batched_datasets)
all_cn_sentences = []
all_en_sentences = []
for features, labels in batched_dataset_features:
    all_en_sentences += [feature.decode('utf-8') for feature in features]
    all_cn_sentences += [label.decode('utf-8') for label in labels]
input_tensor, target_tensor, inp_lang, targ_lang = TranslateDatasetBuilder.load_dataset_from_sentences(all_en_sentences,
                                                                                                       all_cn_sentences)
# TranslateDatasetBuilder.convert(inp_lang, input_tensor[0])
dataset = tf.data.Dataset.from_tensor_slices((input_tensor, target_tensor)).batch(BATCH_SIZE, drop_remainder=True)
train_data, _, test_data = SuperDatasetBuilder.split_full_datasets(dataset, num_examples, train_size=0.9,
                                                                   validation_size=0, test_size=0.1)
###############################
vocab_inp_size = len(inp_lang.word_index) + 1
vocab_tar_size = len(targ_lang.word_index) + 1
###############################
# encoder layer
encoder = Encoder(vocab_inp_size, embedding_dim, units, BATCH_SIZE)
# decoder layer
decoder = Decoder(vocab_tar_size, embedding_dim, units, BATCH_SIZE)
# attention layer
attention_layer = BahdanauAttention(10)
# 定义优化器和损失函数
optimizer = tf.keras.optimizers.Adam()
loss_object = tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True, reduction='none')
# 定义检查点（基于对象保存）
checkpoint_dir = '/home/user/tmp/atttion/training_checkpoints'
checkpoint_prefix = os.path.join(checkpoint_dir, "ckpt")
checkpoint = tf.train.Checkpoint(optimizer=optimizer,
                                 encoder=encoder,
                                 decoder=decoder)


def loss_function(real, pred):
    mask = tf.math.logical_not(tf.math.equal(real, 0))
    loss_ = loss_object(real, pred)
    mask = tf.cast(mask, dtype=loss_.dtype)
    loss_ *= mask
    return tf.reduce_mean(loss_)


@tf.function
def train_step(inp, targ, enc_hidden):
    loss = 0
    with tf.GradientTape() as tape:
        enc_output, enc_hidden = encoder(inp, enc_hidden)
        dec_hidden = enc_hidden
        dec_input = tf.expand_dims([targ_lang.word_index['<start>']] * BATCH_SIZE, 1)
        # 教师强制 - 将目标词作为下一个输入
        for t in range(1, targ.shape[1]):
            # 将编码器输出 （enc_output） 传送至解码器
            predictions, dec_hidden, _ = decoder(dec_input, dec_hidden, enc_output)
            loss += loss_function(targ[:, t], predictions)
            # 使用教师强制
            dec_input = tf.expand_dims(targ[:, t], 1)
    batch_loss = (loss / int(targ.shape[1]))
    variables = encoder.trainable_variables + decoder.trainable_variables
    gradients = tape.gradient(loss, variables)
    optimizer.apply_gradients(zip(gradients, variables))
    return batch_loss


@print_exec_time
def training_attention_model():
    for epoch in range(EPOCHS):
        start = time.time()
        enc_hidden = encoder.initialize_hidden_state()
        total_loss = 0
        for batch, (input_batch, target_batch) in enumerate(tfds.as_numpy(train_data)):
            # 样本输入
            sample_hidden = encoder.initialize_hidden_state()
            # encoder
            sample_output, sample_hidden = encoder(input_batch, sample_hidden)
            print('Encoder output shape: (batch size, sequence length, units) {}'.format(sample_output.shape))
            print('Encoder Hidden state shape: (batch size, units) {}'.format(sample_hidden.shape))
            # attention
            attention_result, attention_weights = attention_layer(sample_hidden, sample_output)
            print("Attention result shape: (batch size, units) {}".format(attention_result.shape))
            print("Attention weights shape: (batch_size, sequence_length, 1) {}".format(attention_weights.shape))
            # decoder
            sample_decoder_output, _, _ = decoder(tf.random.uniform((BATCH_SIZE, 1)), sample_hidden, sample_output)
            print('Decoder output shape: (batch_size, vocab size) {}'.format(sample_decoder_output.shape))
            batch_loss = train_step(input_batch, target_batch, enc_hidden)
            total_loss += batch_loss
            if batch % 10 == 0:
                print('Epoch {} Batch {} Loss {:.4f}'.format(epoch + 1,
                                                             batch,
                                                             batch_loss.numpy()))
        # 每 1 个周期（epoch），保存（检查点）一次模型
        if (epoch + 1) % 1 == 0:
            checkpoint.save(file_prefix=checkpoint_prefix)
        print('Epoch {} Loss {:.4f}'.format(epoch + 1, total_loss))
        print('Time taken for 1 epoch {} sec\n'.format(time.time() - start))


def evaluate(sentence):
    max_length_targ = 38  # from TranslateDatasetBuilder
    max_length_inp = 44  # from TranslateDatasetBuilder
    attention_plot = np.zeros((max_length_targ, max_length_inp))
    sentence = TranslateDatasetBuilder.preprocess_sentence(sentence)
    inputs = [inp_lang.word_index[i] for i in sentence.split(' ')]
    inputs = tf.keras.preprocessing.sequence.pad_sequences([inputs],
                                                           maxlen=max_length_inp,
                                                           padding='post')
    inputs = tf.convert_to_tensor(inputs)
    result = ''
    hidden = [tf.zeros((1, units))]
    enc_out, enc_hidden = encoder(inputs, hidden)
    dec_hidden = enc_hidden
    dec_input = tf.expand_dims([targ_lang.word_index['<start>']], 0)
    for t in range(max_length_targ):
        predictions, dec_hidden, attention_weights = decoder(dec_input,
                                                             dec_hidden,
                                                             enc_out)
        # 存储注意力权重以便后面制图
        attention_weights = tf.reshape(attention_weights, (-1,))
        attention_plot[t] = attention_weights.numpy()
        predicted_id = tf.argmax(predictions[0]).numpy()
        result += targ_lang.index_word[predicted_id] + ' '
        if targ_lang.index_word[predicted_id] == '<end>':
            return result, sentence, attention_plot
        # 预测的 ID 被输送回模型
        dec_input = tf.expand_dims([predicted_id], 0)
    return result, sentence, attention_plot


def translate(sentence):
    result, sentence, attention_plot = evaluate(sentence)
    print(f'Input: {sentence}')
    print(f'Predicted translation: {result}')
    attention_plot = attention_plot[:len(result.split(' ')), :len(sentence.split(' '))]
    plot_attention(attention_plot, sentence.split(' '), result.split(' '))


# 注意力权重制图函数
def plot_attention(attention, sentence, predicted_sentence):
    fig = plt.figure(figsize=(10, 10))
    ax = fig.add_subplot(1, 1, 1)
    ax.matshow(attention, cmap='viridis')
    fontdict = {'fontsize': 14}
    ax.set_xticklabels([''] + sentence, fontdict=fontdict, rotation=90)
    ax.set_yticklabels([''] + predicted_sentence, fontdict=fontdict)
    ax.xaxis.set_major_locator(ticker.MultipleLocator(1))
    ax.yaxis.set_major_locator(ticker.MultipleLocator(1))
    plt.show()


if __name__ == '__main__':
    # training_attention_model()
    # 恢复检查点目录（checkpoint_dir）中最新的检查点
    saved_model = checkpoint.restore(tf.train.latest_checkpoint(checkpoint_dir))  # 可以查看checkpoint.encoder、optimizer、decoder对象的状态
    # 翻译
    translate("今 天 家 里 下 雪 了 哦")




class Portfolio(object):
    '''
    Parameters:
    trading_cost: Cost of taking a position is 3 pips by default
    _prices: Dataframe of Open and Close prices for reward calculation
    reward_normalizer: to convert precision of prices (0.0001) to number (1) for reward calculation
    total_reward: Keep track of reward accumulated in an episode
    current_time: the current index in the dataframe
    step: keep track of number of steps taken. Different from current_time

    curr_trade: A dictionary that records the details of an open position
    journal: Stores all records of curr_trade
    holding_trade: Boolean Flag to allow new position

    '''

    def __init__(self, prices, train_end_index, trade_period=1, max_price=10, denom=0.0001, cost=3):

        self.train_end = train_end_index
        self.trade_period = trade_period

        # Trading cost is 3 pips
        self.trading_cost = cost

        # Store list of Open price and Close price to manage reward calculation
        self._open = prices.Open.values
        self._close = prices.Close.values
        self._index = prices.index

        # To normalise reward terms
        self.reward_normalizer = 1. / denom
        self._reset()
        self.open_trade = False

    def _reset(self, train=True):
        self.total_reward = 0
        self.total_trades = 0
        self.average_profit_per_trade = 0
        self.count_open_trades = 0

        if train:
            self.current_time = 1
        else:
            self.current_time = 1

        self.curr_trade = {'Entry Price': 0, 'Exit Price': 0, 'Entry Time': None, 'Exit Time': None, 'Profit': 0,
                           'Trade Duration': 0, 'Type': None, 'reward': 0}
        self.journal = []
        self.open_trade = False

    def _reset_trade(self):
        self.curr_trade = {'Entry Price': 0, 'Exit Price': 0, 'Entry Time': None, 'Exit Time': None, 'Profit': 0,
                           'Trade Duration': 0, 'Type': None, 'reward': 0}

    def close_trade(self, curr_close_price, curr_time):
        if self.curr_trade['Type'] == 'SELL':
            self.count_open_trades -= 1
            # Update remaining keys in curr_trade dict
            self.curr_trade['Exit Price'] = curr_close_price
            self.curr_trade['Exit Time'] = curr_time
            self.curr_trade['Profit'] = -1 * (
                curr_close_price - self.curr_trade['Entry Price']) * self.reward_normalizer - self.trading_cost
            # Add the current trade to the journal
            self.journal.append(self.curr_trade)
            self._reset_trade()
            self.open_trade = False

        if self.curr_trade['Type'] == 'BUY':
            self.count_open_trades -= 1
            # Action is 1, Selling to close the Long position
            # Update remaining  keys in curr_trade dict
            self.curr_trade['Exit Price'] = curr_close_price
            self.curr_trade['Exit Time'] = curr_time
            self.curr_trade['Profit'] = (curr_close_price - self.curr_trade[
                'Entry Price']) * self.reward_normalizer - self.trading_cost
            # Add curr_trade to journal, then reset curr_trade
            self.journal.append(self.curr_trade)
            self._reset_trade()

            self.open_trade = False

    def _holding_trade(self, curr_close_price, prev_close_price, reward):
        self.curr_trade['Trade Duration'] += 1
        if self.curr_trade['Type'] == 'SELL':
            reward = -1 * (curr_close_price - prev_close_price) * self.reward_normalizer
        if self.curr_trade['Type'] == 'BUY':
            reward = (curr_close_price - prev_close_price) * self.reward_normalizer
        return reward

    def _step(self, action):
        curr_open_price = self._open[self.current_time]
        curr_close_price = self._close[self.current_time]
        curr_time = self._index[self.current_time]
        prev_close_price = self._close[self.current_time - 1]
        reward = 0
        if action == 3 or self.curr_trade['Trade Duration'] >= self.trade_period:
            # Closing trade or trade duration is reached
            self.close_trade(curr_close_price, curr_time)
        elif action == 1:
            if not self.open_trade:
                # BUYING
                self.curr_trade['Entry Price'] = curr_open_price
                self.curr_trade['Type'] = "BUY"
                self.curr_trade['Entry Time'] = curr_time
                self.curr_trade['Trade Duration'] += 1
                reward = (curr_close_price - curr_open_price) * self.reward_normalizer - self.trading_cost
                self.total_trades += 1
                self.open_trade = True
                self.count_open_trades += 1
            else:
                reward = self._holding_trade(curr_close_price, prev_close_price, reward)
        elif action == 2:
            if not self.open_trade:
                # SELLING
                self.curr_trade['Entry Price'] = curr_open_price
                self.curr_trade['Type'] = "SELL"
                self.curr_trade['Entry Time'] = curr_time
                self.curr_trade['Trade Duration'] += 1
                reward = -1 * (curr_close_price - curr_open_price) * self.reward_normalizer - self.trading_cost
                self.total_trades += 1
                self.open_trade = True
                self.count_open_trades += 1
            else:
                reward = self._holding_trade(curr_close_price, prev_close_price, reward)
        elif action == 0:
            # Holding trade
            if self.open_trade:
                reward = self._holding_trade(curr_close_price, prev_close_price, reward)
            else:
                pass
        self.curr_trade["reward"] += reward
        self.total_reward += reward

        if self.total_trades > 0:
            self.average_profit_per_trade = self.total_reward / self.total_trades

        self.current_time += 1

        info = {'Average reward per trade': self.average_profit_per_trade,
                'Reward for this trade': self.curr_trade["reward"],
                'Total reward': self.total_reward}

        return self.curr_trade["reward"], info



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""BERT model input pipelines."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf


def decode_record(record, name_to_features):
  """Decodes a record to a TensorFlow example."""
  example = tf.io.parse_single_example(record, name_to_features)

  # tf.Example only supports tf.int64, but the TPU only supports tf.int32.
  # So cast all int64 to int32.
  for name in list(example.keys()):
    t = example[name]
    if t.dtype == tf.int64:
      t = tf.cast(t, tf.int32)
    example[name] = t

  return example


def single_file_dataset(input_file, name_to_features):
  """Creates a single-file dataset to be passed for BERT custom training."""
  # For training, we want a lot of parallel reading and shuffling.
  # For eval, we want no shuffling and parallel reading doesn't matter.
  d = tf.data.TFRecordDataset(input_file)
  d = d.map(lambda record: decode_record(record, name_to_features))

  # When `input_file` is a path to a single file or a list
  # containing a single path, disable auto sharding so that
  # same input file is sent to all workers.
  if isinstance(input_file, str) or len(input_file) == 1:
    options = tf.data.Options()
    options.experimental_distribute.auto_shard_policy = (
        tf.data.experimental.AutoShardPolicy.OFF)
    d = d.with_options(options)
  return d


def create_pretrain_dataset(input_patterns,
                            seq_length,
                            max_predictions_per_seq,
                            batch_size,
                            is_training=True,
                            input_pipeline_context=None):
  """Creates input dataset from (tf)records files for pretraining."""
  name_to_features = {
      'input_ids':
          tf.io.FixedLenFeature([seq_length], tf.int64),
      'input_mask':
          tf.io.FixedLenFeature([seq_length], tf.int64),
      'segment_ids':
          tf.io.FixedLenFeature([seq_length], tf.int64),
      'masked_lm_positions':
          tf.io.FixedLenFeature([max_predictions_per_seq], tf.int64),
      'masked_lm_ids':
          tf.io.FixedLenFeature([max_predictions_per_seq], tf.int64),
      'masked_lm_weights':
          tf.io.FixedLenFeature([max_predictions_per_seq], tf.float32),
      'next_sentence_labels':
          tf.io.FixedLenFeature([1], tf.int64),
  }

  dataset = tf.data.Dataset.list_files(input_patterns, shuffle=is_training)

  if input_pipeline_context and input_pipeline_context.num_input_pipelines > 1:
    dataset = dataset.shard(input_pipeline_context.num_input_pipelines,
                            input_pipeline_context.input_pipeline_id)

  dataset = dataset.repeat()

  # We set shuffle buffer to exactly match total number of
  # training files to ensure that training data is well shuffled.
  input_files = []
  for input_pattern in input_patterns:
    input_files.extend(tf.io.gfile.glob(input_pattern))
  dataset = dataset.shuffle(len(input_files))

  # In parallel, create tf record dataset for each train files.
  # cycle_length = 8 means that up to 8 files will be read and deserialized in
  # parallel. You may want to increase this number if you have a large number of
  # CPU cores.
  dataset = dataset.interleave(
      tf.data.TFRecordDataset, cycle_length=8,
      num_parallel_calls=tf.data.experimental.AUTOTUNE)

  decode_fn = lambda record: decode_record(record, name_to_features)
  dataset = dataset.map(
      decode_fn, num_parallel_calls=tf.data.experimental.AUTOTUNE)

  def _select_data_from_record(record):
    """Filter out features to use for pretraining."""
    x = {
        'input_word_ids': record['input_ids'],
        'input_mask': record['input_mask'],
        'input_type_ids': record['segment_ids'],
        'masked_lm_positions': record['masked_lm_positions'],
        'masked_lm_ids': record['masked_lm_ids'],
        'masked_lm_weights': record['masked_lm_weights'],
        'next_sentence_labels': record['next_sentence_labels'],
    }

    y = record['masked_lm_weights']

    return (x, y)

  dataset = dataset.map(
      _select_data_from_record,
      num_parallel_calls=tf.data.experimental.AUTOTUNE)

  if is_training:
    dataset = dataset.shuffle(100)

  dataset = dataset.batch(batch_size, drop_remainder=True)
  dataset = dataset.prefetch(1024)
  return dataset


def create_classifier_dataset(file_path,
                              seq_length,
                              batch_size,
                              is_training=True,
                              input_pipeline_context=None):
  """Creates input dataset from (tf)records files for train/eval."""
  name_to_features = {
      'input_ids': tf.io.FixedLenFeature([seq_length], tf.int64),
      'input_mask': tf.io.FixedLenFeature([seq_length], tf.int64),
      'segment_ids': tf.io.FixedLenFeature([seq_length], tf.int64),
      'label_ids': tf.io.FixedLenFeature([], tf.int64),
      'is_real_example': tf.io.FixedLenFeature([], tf.int64),
  }
  dataset = single_file_dataset(file_path, name_to_features)

  # The dataset is always sharded by number of hosts.
  # num_input_pipelines is the number of hosts rather than number of cores.
  if input_pipeline_context and input_pipeline_context.num_input_pipelines > 1:
    dataset = dataset.shard(input_pipeline_context.num_input_pipelines,
                            input_pipeline_context.input_pipeline_id)

  def _select_data_from_record(record):
    x = {
        'input_word_ids': record['input_ids'],
        'input_mask': record['input_mask'],
        'input_type_ids': record['segment_ids']
    }
    y = record['label_ids']
    return (x, y)

  dataset = dataset.map(_select_data_from_record)

  if is_training:
    dataset = dataset.shuffle(100)
    dataset = dataset.repeat()

  dataset = dataset.batch(batch_size, drop_remainder=is_training)
  dataset = dataset.prefetch(1024)
  return dataset


def create_squad_dataset(file_path,
                         seq_length,
                         batch_size,
                         is_training=True,
                         input_pipeline_context=None):
  """Creates input dataset from (tf)records files for train/eval."""
  name_to_features = {
      'input_ids': tf.io.FixedLenFeature([seq_length], tf.int64),
      'input_mask': tf.io.FixedLenFeature([seq_length], tf.int64),
      'segment_ids': tf.io.FixedLenFeature([seq_length], tf.int64),
  }
  if is_training:
    name_to_features['start_positions'] = tf.io.FixedLenFeature([], tf.int64)
    name_to_features['end_positions'] = tf.io.FixedLenFeature([], tf.int64)
  else:
    name_to_features['unique_ids'] = tf.io.FixedLenFeature([], tf.int64)

  dataset = single_file_dataset(file_path, name_to_features)

  # The dataset is always sharded by number of hosts.
  # num_input_pipelines is the number of hosts rather than number of cores.
  if input_pipeline_context and input_pipeline_context.num_input_pipelines > 1:
    dataset = dataset.shard(input_pipeline_context.num_input_pipelines,
                            input_pipeline_context.input_pipeline_id)

  def _select_data_from_record(record):
    """Dispatches record to features and labels."""
    x, y = {}, {}
    for name, tensor in record.items():
      if name in ('start_positions', 'end_positions'):
        y[name] = tensor
      elif name == 'input_ids':
        x['input_word_ids'] = tensor
      elif name == 'segment_ids':
        x['input_type_ids'] = tensor
      else:
        x[name] = tensor
    return (x, y)

  dataset = dataset.map(_select_data_from_record)

  if is_training:
    dataset = dataset.shuffle(100)
    dataset = dataset.repeat()

  dataset = dataset.batch(batch_size, drop_remainder=True)
  dataset = dataset.prefetch(1024)
  return dataset



import logging
import pickle
import sys
import time

import networkx as nx
import ray
import skydl.ray.streaming_0_8_2.python.processor as processor
import skydl.ray.streaming_0_8_2.python.runtime.transfer as transfer
from skydl.ray.streaming_0_8_2.python.communication import DataChannel
from skydl.ray.streaming_0_8_2.python.config import Config
from skydl.ray.streaming_0_8_2.python.jobworker import JobWorker
from skydl.ray.streaming_0_8_2.python.operator import Operator, OpType
from skydl.ray.streaming_0_8_2.python.operator import PScheme, PStrategy

logger = logging.getLogger(__name__)
logger.setLevel("INFO")


# Rolling sum's logic
def _sum(value_1, value_2):
    return value_1 + value_2


# Partitioning strategies that require all-to-all instance communication
all_to_all_strategies = [
    PStrategy.Shuffle, PStrategy.ShuffleByKey, PStrategy.Broadcast,
    PStrategy.RoundRobin
]


# Environment configuration
class Conf:
    """Environment configuration.

    This class includes all information about the configuration of the
    streaming environment.
    """

    def __init__(self, parallelism=1, channel_type=Config.MEMORY_CHANNEL):
        self.parallelism = parallelism
        self.channel_type = channel_type
        # ...


class ExecutionGraph:
    def __init__(self, env):
        self.env = env
        self.physical_topo = nx.DiGraph()  # DAG
        # Handles to all actors in the physical dataflow
        self.actor_handles = []
        # (op_id, op_instance_index) -> ActorID
        self.actors_map = {}
        # execution graph build time: milliseconds since epoch
        self.build_time = 0
        self.task_id_counter = 0
        self.task_ids = {}
        self.input_channels = {}  # operator id -> input channels
        self.output_channels = {}  # operator id -> output channels

    # Constructs and deploys a Ray actor of a specific type
    # TODO (john): Actor placement information should be specified in
    # the environment's configuration
    def __generate_actor(self, instance_index, operator, input_channels,
                         output_channels):
        """Generates an actor that will execute a particular instance of
        the logical operator

        Attributes:
            instance_index: The index of the instance the actor will execute.
            operator: The metadata of the logical operator.
            input_channels: The input channels of the instance.
            output_channels The output channels of the instance.
        """
        worker_id = (operator.id, instance_index)
        # Record the physical dataflow graph (for debugging purposes)
        self.__add_channel(worker_id, output_channels)
        # Note direct_call only support pass by value
        return JobWorker._remote(
            args=[worker_id, operator, input_channels, output_channels],
            is_direct_call=True)

    # Constructs and deploys a Ray actor for each instance of
    # the given operator
    def __generate_actors(self, operator, upstream_channels,
                          downstream_channels):
        """Generates one actor for each instance of the given logical
        operator.

        Attributes:
            operator (Operator): The logical operator metadata.
            upstream_channels (list): A list of all upstream channels for
            all instances of the operator.
            downstream_channels (list): A list of all downstream channels
            for all instances of the operator.
        """
        num_instances = operator.num_instances
        logger.info("Generating {} actors of type {}...".format(
            num_instances, operator.type))
        handles = []
        for i in range(num_instances):
            # Collect input and output channels for the particular instance
            ip = [c for c in upstream_channels if c.dst_instance_index == i]
            op = [c for c in downstream_channels if c.src_instance_index == i]
            log = "Constructed {} input and {} output channels "
            log += "for the {}-th instance of the {} operator."
            logger.debug(log.format(len(ip), len(op), i, operator.type))
            handle = self.__generate_actor(i, operator, ip, op)
            if handle:
                handles.append(handle)
                self.actors_map[(operator.id, i)] = handle
        return handles

    # Adds a channel/edge to the physical dataflow graph
    def __add_channel(self, actor_id, output_channels):
        for c in output_channels:
            dest_actor_id = (c.dst_operator_id, c.dst_instance_index)
            self.physical_topo.add_edge(actor_id, dest_actor_id)

    # Generates all required data channels between an operator
    # and its downstream operators
    def _generate_channels(self, operator):
        """Generates all output data channels
        (see: DataChannel in communication.py) for all instances of
        the given logical operator.

        The function constructs one data channel for each pair of
        communicating operator instances (instance_1,instance_2),
        where instance_1 is an instance of the given operator and instance_2
        is an instance of a direct downstream operator.

        The number of total channels generated depends on the partitioning
        strategy specified by the user.
        """
        channels = {}  # destination operator id -> channels
        strategies = operator.partitioning_strategies
        for dst_operator, p_scheme in strategies.items():
            num_dest_instances = self.env.operators[dst_operator].num_instances
            entry = channels.setdefault(dst_operator, [])
            if p_scheme.strategy == PStrategy.Forward:
                for i in range(operator.num_instances):
                    # ID of destination instance to connect
                    id = i % num_dest_instances
                    qid = self._gen_str_qid(operator.id, i, dst_operator, id)
                    c = DataChannel(operator.id, i, dst_operator, id, qid)
                    entry.append(c)
            elif p_scheme.strategy in all_to_all_strategies:
                for i in range(operator.num_instances):
                    for j in range(num_dest_instances):
                        qid = self._gen_str_qid(operator.id, i, dst_operator,
                                                j)
                        c = DataChannel(operator.id, i, dst_operator, j, qid)
                        entry.append(c)
            else:
                # TODO (john): Add support for other partitioning strategies
                sys.exit("Unrecognized or unsupported partitioning strategy.")
        return channels

    def _gen_str_qid(self, src_operator_id, src_instance_index,
                     dst_operator_id, dst_instance_index):
        from_task_id = self.env.execution_graph.get_task_id(
            src_operator_id, src_instance_index)
        to_task_id = self.env.execution_graph.get_task_id(
            dst_operator_id, dst_instance_index)
        return transfer.ChannelID.gen_id(from_task_id, to_task_id,
                                         self.build_time)

    def _gen_task_id(self):
        task_id = self.task_id_counter
        self.task_id_counter += 1
        return task_id

    def get_task_id(self, op_id, op_instance_id):
        return self.task_ids[(op_id, op_instance_id)]

    def get_actor(self, op_id, op_instance_id):
        return self.actors_map[(op_id, op_instance_id)]

    # Prints the physical dataflow graph
    def print_physical_graph(self):
        logger.info("===================================")
        logger.info("======Physical Dataflow Graph======")
        logger.info("===================================")
        # Print all data channels between operator instances
        log = "(Source Operator ID,Source Operator Name,Source Instance ID)"
        log += " --> "
        log += "(Destination Operator ID,Destination Operator Name,"
        log += "Destination Instance ID)"
        logger.info(log)
        for src_actor_id, dst_actor_id in self.physical_topo.edges:
            src_operator_id, src_instance_index = src_actor_id
            dst_operator_id, dst_instance_index = dst_actor_id
            logger.info("({},{},{}) --> ({},{},{})".format(
                src_operator_id, self.env.operators[src_operator_id].name,
                src_instance_index, dst_operator_id,
                self.env.operators[dst_operator_id].name, dst_instance_index))

    def build_graph(self):
        self.build_channels()

        # to support cyclic reference serialization
        try:
            ray.register_custom_serializer(Environment, use_pickle=True)
            ray.register_custom_serializer(ExecutionGraph, use_pickle=True)
            ray.register_custom_serializer(OpType, use_pickle=True)
            ray.register_custom_serializer(PStrategy, use_pickle=True)
        except Exception:
            # local mode can't use pickle
            pass

        # Each operator instance is implemented as a Ray actor
        # Actors are deployed in topological order, as we traverse the
        # logical dataflow from sources to sinks.
        for node in nx.topological_sort(self.env.logical_topo):
            operator = self.env.operators[node]
            # Instantiate Ray actors
            handles = self.__generate_actors(
                operator, self.input_channels.get(node, []),
                self.output_channels.get(node, []))
            if handles:
                self.actor_handles.extend(handles)

    def build_channels(self):
        self.build_time = int(time.time() * 1000)
        # gen auto-incremented unique task id for every operator instance
        for node in nx.topological_sort(self.env.logical_topo):
            operator = self.env.operators[node]
            for i in range(operator.num_instances):
                operator_instance_id = (operator.id, i)
                self.task_ids[operator_instance_id] = self._gen_task_id()
        channels = {}
        for node in nx.topological_sort(self.env.logical_topo):
            operator = self.env.operators[node]
            # Generate downstream data channels
            downstream_channels = self._generate_channels(operator)
            channels[node] = downstream_channels
        # op_id -> channels
        input_channels = {}
        output_channels = {}
        for op_id, all_downstream_channels in channels.items():
            for dst_op_channels in all_downstream_channels.values():
                for c in dst_op_channels:
                    dst = input_channels.setdefault(c.dst_operator_id, [])
                    dst.append(c)
                    src = output_channels.setdefault(c.src_operator_id, [])
                    src.append(c)
        self.input_channels = input_channels
        self.output_channels = output_channels


# The execution environment for a streaming job
class Environment:
    """A streaming environment.

    This class is responsible for constructing the logical and the
    physical dataflow.

    Attributes:
         logical_topo (DiGraph): The user-defined logical topology in
         NetworkX DiGRaph format.
         (See: https://networkx.github.io)
         physical_topo (DiGraph): The physical topology in NetworkX
         DiGRaph format. The physical dataflow is constructed by the
         environment based on logical_topo.
         operators (dict): A mapping from operator ids to operator metadata
         (See: Operator in operator.py).
         config (Config): The environment's configuration.
         topo_cleaned (bool): A flag that indicates whether the logical
         topology is garbage collected (True) or not (False).
         actor_handles (list): A list of all Ray actor handles that execute
         the streaming dataflow.
    """

    def __init__(self, config=Conf()):
        self.logical_topo = nx.DiGraph()  # DAG
        self.operators = {}  # operator id --> operator object
        self.config = config  # Environment's configuration
        self.topo_cleaned = False
        self.operator_id_counter = 0
        self.execution_graph = None  # set when executed

    def gen_operator_id(self):
        op_id = self.operator_id_counter
        self.operator_id_counter += 1
        return op_id

    # An edge denotes a flow of data between logical operators
    # and may correspond to multiple data channels in the physical dataflow
    def _add_edge(self, source, destination):
        self.logical_topo.add_edge(source, destination)

    # Cleans the logical dataflow graph to construct and
    # deploy the physical dataflow
    def _collect_garbage(self):
        if self.topo_cleaned is True:
            return
        for node in self.logical_topo:
            self.operators[node]._clean()
        self.topo_cleaned = True

    # Sets the level of parallelism for a registered operator
    # Overwrites the environment parallelism (if set)
    def _set_parallelism(self, operator_id, level_of_parallelism):
        self.operators[operator_id].num_instances = level_of_parallelism

    # Sets the same level of parallelism for all operators in the environment
    def set_parallelism(self, parallelism):
        self.config.parallelism = parallelism

    # Creates and registers a user-defined data source
    # TODO (john): There should be different types of sources, e.g. sources
    # reading from Kafka, text files, etc.
    # TODO (john): Handle case where environment parallelism is set
    def source(self, source):
        source_id = self.gen_operator_id()
        source_stream = DataStream(self, source_id)
        self.operators[source_id] = Operator(
            source_id, OpType.Source, processor.Source, "Source", logic=source)
        return source_stream

    # Creates and registers a new data source that reads a
    # text file line by line
    # TODO (john): There should be different types of sources,
    # e.g. sources reading from Kafka, text files, etc.
    # TODO (john): Handle case where environment parallelism is set
    def read_text_file(self, filepath):
        source_id = self.gen_operator_id()
        source_stream = DataStream(self, source_id)
        self.operators[source_id] = Operator(
            source_id,
            OpType.ReadTextFile,
            processor.ReadTextFile,
            "Read Text File",
            other=filepath)
        return source_stream

    # Constructs and deploys the physical dataflow
    def execute(self):
        """Deploys and executes the physical dataflow."""
        self._collect_garbage()  # Make sure everything is clean
        # TODO (john): Check if dataflow has any 'logical inconsistencies'
        # For example, if there is a forward partitioning strategy but
        # the number of downstream instances is larger than the number of
        # upstream instances, some of the downstream instances will not be
        # used at all

        self.execution_graph = ExecutionGraph(self)
        self.execution_graph.build_graph()
        logger.info("init...")
        # init
        init_waits = []
        for actor_handle in self.execution_graph.actor_handles:
            init_waits.append(actor_handle.init.remote(pickle.dumps(self)))
        for wait in init_waits:
            assert ray.get(wait) is True
        logger.info("running...")
        # start
        exec_handles = []
        for actor_handle in self.execution_graph.actor_handles:
            exec_handles.append(actor_handle.start.remote())

        return exec_handles

    def wait_finish(self):
        for actor_handle in self.execution_graph.actor_handles:
            while not ray.get(actor_handle.is_finished.remote()):
                time.sleep(1)

    # Prints the logical dataflow graph
    def print_logical_graph(self):
        self._collect_garbage()
        logger.info("==================================")
        logger.info("======Logical Dataflow Graph======")
        logger.info("==================================")
        # Print operators in topological order
        for node in nx.topological_sort(self.logical_topo):
            downstream_neighbors = list(self.logical_topo.neighbors(node))
            logger.info("======Current Operator======")
            operator = self.operators[node]
            operator.print()
            logger.info("======Downstream Operators======")
            if len(downstream_neighbors) == 0:
                logger.info("None\n")
            for downstream_node in downstream_neighbors:
                self.operators[downstream_node].print()


# TODO (john): We also need KeyedDataStream and WindowedDataStream as
# subclasses of DataStream to prevent ill-defined logical dataflows


# A DataStream corresponds to an edge in the logical dataflow
class DataStream:
    """A data stream.

    This class contains all information about a logical stream, i.e. an edge
    in the logical topology. It is the main class exposed to the user.

    Attributes:
         id (UUID): The id of the stream
         env (Environment): The environment the stream belongs to.
         src_operator_id (UUID): The id of the source operator of the stream.
         dst_operator_id (UUID): The id of the destination operator of the
         stream.
         is_partitioned (bool): Denotes if there is a partitioning strategy
         (e.g. shuffle) for the stream or not (default stategy: Forward).
    """
    stream_id_counter = 0

    def __init__(self,
                 environment,
                 source_id=None,
                 dest_id=None,
                 is_partitioned=False):
        self.env = environment
        self.id = DataStream.stream_id_counter
        DataStream.stream_id_counter += 1
        self.src_operator_id = source_id
        self.dst_operator_id = dest_id
        # True if a partitioning strategy for this stream exists,
        # false otherwise
        self.is_partitioned = is_partitioned

    # Generates a new stream after a data transformation is applied
    def __expand(self):
        stream = DataStream(self.env)
        assert (self.dst_operator_id is not None)
        stream.src_operator_id = self.dst_operator_id
        stream.dst_operator_id = None
        return stream

    # Assigns the partitioning strategy to a new 'open-ended' stream
    # and returns the stream. At this point, the partitioning strategy
    # is not associated with any destination operator. We expect this to
    # be done later, as we continue assembling the dataflow graph
    def __partition(self, strategy, partition_fn=None):
        scheme = PScheme(strategy, partition_fn)
        source_operator = self.env.operators[self.src_operator_id]
        new_stream = DataStream(
            self.env, source_id=source_operator.id, is_partitioned=True)
        source_operator._set_partition_strategy(new_stream.id, scheme)
        return new_stream

    # Registers the operator to the environment and returns a new
    # 'open-ended' stream. The registered operator serves as the destination
    # of the previously 'open' stream
    def __register(self, operator):
        """Registers the given logical operator to the environment and
        connects it to its upstream operator (if any).

        A call to this function adds a new edge to the logical topology.

        Attributes:
             operator (Operator): The metadata of the logical operator.
        """
        self.env.operators[operator.id] = operator
        self.dst_operator_id = operator.id
        logger.debug("Adding new dataflow edge ({},{}) --> ({},{})".format(
            self.src_operator_id,
            self.env.operators[self.src_operator_id].name,
            self.dst_operator_id,
            self.env.operators[self.dst_operator_id].name))
        # Update logical dataflow graphs
        self.env._add_edge(self.src_operator_id, self.dst_operator_id)
        # Keep track of the partitioning strategy and the destination operator
        src_operator = self.env.operators[self.src_operator_id]
        if self.is_partitioned is True:
            partitioning, _ = src_operator._get_partition_strategy(self.id)
            src_operator._set_partition_strategy(self.id, partitioning,
                                                 operator.id)
        elif src_operator.type == OpType.KeyBy:
            # Set the output partitioning strategy to shuffle by key
            partitioning = PScheme(PStrategy.ShuffleByKey)
            src_operator._set_partition_strategy(self.id, partitioning,
                                                 operator.id)
        else:  # No partitioning strategy has been defined - set default
            partitioning = PScheme(PStrategy.Forward)
            src_operator._set_partition_strategy(self.id, partitioning,
                                                 operator.id)
        return self.__expand()

    # Sets the level of parallelism for an operator, i.e. its total
    # number of instances. Each operator instance corresponds to an actor
    # in the physical dataflow
    def set_parallelism(self, num_instances):
        """Sets the number of instances for the source operator of the stream.

        Attributes:
             num_instances (int): The level of parallelism for the source
             operator of the stream.
        """
        assert (num_instances > 0)
        self.env._set_parallelism(self.src_operator_id, num_instances)
        return self

    # Stream Partitioning Strategies #
    # TODO (john): Currently, only forward (default), shuffle,
    # and broadcast are supported

    # Hash-based record shuffling
    def shuffle(self):
        """Registers a shuffling partitioning strategy for the stream."""
        return self.__partition(PStrategy.Shuffle)

    # Broadcasts each record to all downstream instances
    def broadcast(self):
        """Registers a broadcast partitioning strategy for the stream."""
        return self.__partition(PStrategy.Broadcast)

    # Rescales load to downstream instances
    def rescale(self):
        """Registers a rescale partitioning strategy for the stream.

        Same as Flink's rescale (see: https://ci.apache.org/projects/flink/
        flink-docs-stable/dev/stream/operators/#physical-partitioning).
        """
        return self.__partition(PStrategy.Rescale)

    # Round-robin partitioning
    def round_robin(self):
        """Registers a round-robin partitioning strategy for the stream."""
        return self.__partition(PStrategy.RoundRobin)

    # User-defined partitioning
    def partition(self, partition_fn):
        """Registers a user-defined partitioning strategy for the stream.

        Attributes:
             partition_fn (function): The user-defined partitioning function.
        """
        return self.__partition(PStrategy.Custom, partition_fn)

    #   Data Trasnformations   #
    # TODO (john): Expand set of supported operators.
    # TODO (john): To support event-time windows we need a mechanism for
    # generating and processing watermarks

    # Registers map operator to the environment
    def map(self, map_fn, name="Map"):
        """Applies a map operator to the stream.

        Attributes:
             map_fn (function): The user-defined logic of the map.
        """
        op = Operator(
            self.env.gen_operator_id(),
            OpType.Map,
            processor.Map,
            name,
            map_fn,
            num_instances=self.env.config.parallelism)
        return self.__register(op)

    # Registers flatmap operator to the environment
    def flat_map(self, flatmap_fn):
        """Applies a flatmap operator to the stream.

        Attributes:
             flatmap_fn (function): The user-defined logic of the flatmap
             (e.g. split()).
        """
        op = Operator(
            self.env.gen_operator_id(),
            OpType.FlatMap,
            processor.FlatMap,
            "FlatMap",
            flatmap_fn,
            num_instances=self.env.config.parallelism)
        return self.__register(op)

    # Registers keyBy operator to the environment
    # TODO (john): This should returned a KeyedDataStream
    def key_by(self, key_selector):
        """Applies a key_by operator to the stream.

        Attributes:
             key_attribute_index (int): The index of the key attributed
             (assuming tuple records).
        """
        op = Operator(
            self.env.gen_operator_id(),
            OpType.KeyBy,
            processor.KeyBy,
            "KeyBy",
            other=key_selector,
            num_instances=self.env.config.parallelism)
        return self.__register(op)

    # Registers Reduce operator to the environment
    def reduce(self, reduce_fn):
        """Applies a rolling sum operator to the stream.

        Attributes:
             sum_attribute_index (int): The index of the attribute to sum
             (assuming tuple records).
        """
        op = Operator(
            self.env.gen_operator_id(),
            OpType.Reduce,
            processor.Reduce,
            "Sum",
            reduce_fn,
            num_instances=self.env.config.parallelism)
        return self.__register(op)

    # Registers Sum operator to the environment
    def sum(self, attribute_selector, state_keeper=None):
        """Applies a rolling sum operator to the stream.

        Attributes:
             sum_attribute_index (int): The index of the attribute to sum
             (assuming tuple records).
        """
        op = Operator(
            self.env.gen_operator_id(),
            OpType.Sum,
            processor.Reduce,
            "Sum",
            _sum,
            other=attribute_selector,
            state_actor=state_keeper,
            num_instances=self.env.config.parallelism)
        return self.__register(op)

    # Registers window operator to the environment.
    # This is a system time window
    # TODO (john): This should return a WindowedDataStream
    def time_window(self, window_width_ms):
        """Applies a system time window to the stream.

        Attributes:
             window_width_ms (int): The length of the window in ms.
        """
        raise Exception("time_window is unsupported")

    # Registers filter operator to the environment
    def filter(self, filter_fn):
        """Applies a filter to the stream.

        Attributes:
             filter_fn (function): The user-defined filter function.
        """
        op = Operator(
            self.env.gen_operator_id(),
            OpType.Filter,
            processor.Filter,
            "Filter",
            filter_fn,
            num_instances=self.env.config.parallelism)
        return self.__register(op)

    # TODO (john): Registers window join operator to the environment
    def window_join(self, other_stream, join_attribute, window_width):
        op = Operator(
            self.env.gen_operator_id(),
            OpType.WindowJoin,
            processor.WindowJoin,
            "WindowJoin",
            num_instances=self.env.config.parallelism)
        return self.__register(op)

    # Registers inspect operator to the environment
    def inspect(self, inspect_logic):
        """Inspects the content of the stream.

        Attributes:
             inspect_logic (function): The user-defined inspect function.
        """
        op = Operator(
            self.env.gen_operator_id(),
            OpType.Inspect,
            processor.Inspect,
            "Inspect",
            inspect_logic,
            num_instances=self.env.config.parallelism)
        return self.__register(op)

    # Registers sink operator to the environment
    # TODO (john): A sink now just drops records but it should be able to
    # export data to other systems
    def sink(self):
        """Closes the stream with a sink operator."""
        op = Operator(
            self.env.gen_operator_id(),
            OpType.Sink,
            processor.Sink,
            "Sink",
            num_instances=self.env.config.parallelism)
        return self.__register(op)



import logging
import pickle
import threading

import ray
import ray.streaming._streaming as _streaming
from skydl.ray.streaming_0_8_2.python.config import Config
from ray._raylet import PythonFunctionDescriptor
from skydl.ray.streaming_0_8_2.python.communication import DataInput, DataOutput

logger = logging.getLogger(__name__)


@ray.remote
class JobWorker:
    """A streaming job worker.

    Attributes:
        worker_id: The id of the instance.
        input_channels: The input gate that manages input channels of
        the instance (see: DataInput in communication.py).
        output_channels (DataOutput): The output gate that manages output
         channels of the instance (see: DataOutput in communication.py).
        the operator instance.
    """

    def __init__(self, worker_id, operator, input_channels, output_channels):
        self.env = None
        self.worker_id = worker_id
        self.operator = operator
        processor_name = operator.processor_class.__name__
        processor_instance = operator.processor_class(operator)
        self.processor_name = processor_name
        self.processor_instance = processor_instance
        self.input_channels = input_channels
        self.output_channels = output_channels
        self.input_gate = None
        self.output_gate = None
        self.reader_client = None
        self.writer_client = None

    def init(self, env):
        """init streaming actor"""
        env = pickle.loads(env)
        self.env = env
        logger.info("init operator instance %s", self.processor_name)

        if env.config.channel_type == Config.NATIVE_CHANNEL:
            core_worker = ray.worker.global_worker.core_worker
            reader_async_func = PythonFunctionDescriptor(
                __name__, self.on_reader_message.__name__,
                self.__class__.__name__)
            reader_sync_func = PythonFunctionDescriptor(
                __name__, self.on_reader_message_sync.__name__,
                self.__class__.__name__)
            self.reader_client = _streaming.ReaderClient(
                core_worker, reader_async_func, reader_sync_func)
            writer_async_func = PythonFunctionDescriptor(
                __name__, self.on_writer_message.__name__,
                self.__class__.__name__)
            writer_sync_func = PythonFunctionDescriptor(
                __name__, self.on_writer_message_sync.__name__,
                self.__class__.__name__)
            self.writer_client = _streaming.WriterClient(
                core_worker, writer_async_func, writer_sync_func)
        if len(self.input_channels) > 0:
            self.input_gate = DataInput(env, self.input_channels)
            self.input_gate.init()
        if len(self.output_channels) > 0:
            self.output_gate = DataOutput(
                env, self.output_channels,
                self.operator.partitioning_strategies)
            self.output_gate.init()
        logger.info("init operator instance %s succeed", self.processor_name)
        return True

    # Starts the actor
    def start(self):
        self.t = threading.Thread(target=self.run, daemon=True)
        self.t.start()
        actor_id = ray.worker.global_worker.actor_id
        logger.info("%s %s started, actor id %s", self.__class__.__name__,
                    self.processor_name, actor_id)

    def run(self):
        logger.info("%s start running", self.processor_name)
        self.processor_instance.run(self.input_gate, self.output_gate)
        logger.info("%s finished running", self.processor_name)
        self.close()

    def close(self):
        if self.input_gate:
            self.input_gate.close()
        if self.output_gate:
            self.output_gate.close()

    def is_finished(self):
        return not self.t.is_alive()

    def on_reader_message(self, buffer: bytes):
        """used in direct call mode"""
        self.reader_client.on_reader_message(buffer)

    def on_reader_message_sync(self, buffer: bytes):
        """used in direct call mode"""
        if self.reader_client is None:
            return b" " * 4  # special flag to indicate this actor not ready
        result = self.reader_client.on_reader_message_sync(buffer)
        return result.to_pybytes()

    def on_writer_message(self, buffer: bytes):
        """used in direct call mode"""
        self.writer_client.on_writer_message(buffer)

    def on_writer_message_sync(self, buffer: bytes):
        """used in direct call mode"""
        if self.writer_client is None:
            return b" " * 4  # special flag to indicate this actor not ready
        result = self.writer_client.on_writer_message_sync(buffer)
        return result.to_pybytes()
